from __future__ import annotations

from typing import Any

from qcfractal.components.record_db_models import OutputStoreORM
from qcportal.compression import compress, CompressionEnum
from qcportal.record_models import OutputTypeEnum


def create_output_orm(output_type: OutputTypeEnum, output: Any) -> OutputStoreORM:
    compressed_out, compression_type, compression_level = compress(output, CompressionEnum.zstd)
    out_orm = OutputStoreORM(
        output_type=output_type,
        compression_type=compression_type,
        compression_level=compression_level,
        data=compressed_out,
    )
    return out_orm



from __future__ import annotations

import copy
import json
import logging
from typing import List, Dict, Tuple, Optional, Sequence, Any, Union, Set, TYPE_CHECKING

import numpy as np
import sqlalchemy.orm.attributes

try:
    from pydantic.v1 import BaseModel, Extra, parse_obj_as
except ImportError:
    from pydantic import BaseModel, Extra, parse_obj_as
from sqlalchemy import select, func
from sqlalchemy.orm import lazyload, joinedload, selectinload, undefer, defer

from qcfractal import __version__ as qcfractal_version
from qcfractal.components.optimization.record_db_models import OptimizationSpecificationORM
from qcfractal.components.services.db_models import ServiceQueueORM, ServiceDependencyORM
from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM
from qcfractal.db_socket.helpers import insert_general
from qcportal.exceptions import MissingDataError
from qcportal.gridoptimization import (
    serialize_key,
    deserialize_key,
    ScanDimension,
    StepTypeEnum,
    GridoptimizationSpecification,
    GridoptimizationQueryFilters,
    GridoptimizationInput,
    GridoptimizationMultiInput,
    GridoptimizationRecord,
)
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.molecules import Molecule
from qcportal.optimization import OptimizationSpecification
from qcportal.record_models import PriorityEnum, RecordStatusEnum, OutputTypeEnum
from qcportal.utils import hash_dict, is_included
from .record_db_models import (
    GridoptimizationSpecificationORM,
    GridoptimizationOptimizationORM,
    GridoptimizationRecordORM,
)
from ..base_record_socket import BaseRecordSocket
from ..record_utils import append_output

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket

# Meaningless, but unique to gridoptimizations
gridoptimization_insert_lock_id = 14300


def expand_ndimensional_grid(
    dimensions: Tuple[int, ...], seeds: Set[Tuple[int, ...]], complete: Set[Tuple[int, ...]]
) -> List[Tuple[Tuple[int, ...], Tuple[int, ...]]]:
    """
    Expands an n-dimensional key/value grid used by gridoptimizations.
    """

    dimensions = tuple(dimensions)
    compute = set()
    connections = []

    for d in range(len(dimensions)):
        # Loop over all compute seeds
        for seed in seeds:
            # Iterate both directions
            for disp in [-1, 1]:
                new_dim = seed[d] + disp

                # Bound check
                if new_dim >= dimensions[d]:
                    continue
                if new_dim < 0:
                    continue

                new = list(seed)
                new[d] = new_dim
                new = tuple(new)

                # Push out duplicates from both new compute and copmlete
                if new in compute:
                    continue
                if new in complete:
                    continue

                compute |= {new}
                connections.append((seed, new))

    return connections


def calculate_starting_grid(scans_dict: Sequence[Dict[str, Any]], molecule: Molecule) -> List[int]:
    """
    Compute the starting parameters for a gridoptimization

    This finds the indices of the steps that most closely matches the given molecule,
    and therefore is a good starting point for the grid optimization

    Parameters
    ----------
    scans_dict
        Information about the scans as a dictionary
    molecule
        Molecule to use for the first grid computations

    Returns
    -------
    :
        Indices of the starting optimization constraints
    """

    scans = parse_obj_as(List[ScanDimension], scans_dict)
    starting_grid = []
    for scan in scans:
        # Find closest index
        if scan.step_type == StepTypeEnum.absolute:
            m = molecule.measure(scan.indices)
        elif scan.step_type == StepTypeEnum.relative:
            m = 0
        else:
            raise KeyError("'step_type' of '{}' not understood.".format(scan.step_type))

        idx = np.abs(np.array(scan.steps) - m).argmin()
        starting_grid.append(int(idx))  # converts from numpy int type

    return starting_grid


class GridoptimizationServiceState(BaseModel):
    """
    This represents the current state of a gridoptimization service
    """

    class Config:
        extra = Extra.forbid
        allow_mutation = True
        validate_assignment = True

    iteration: int
    complete: List[Union[str, Tuple[int, ...]]]
    dimensions: Tuple

    # These are stored as JSON (ie, dict encoded into a string)
    # This makes for faster loads and makes them somewhat tamper-proof
    constraint_template: str


class GridoptimizationRecordSocket(BaseRecordSocket):
    """
    Socket for handling gridoptimization computations
    """

    # Used by the base class
    record_orm = GridoptimizationRecordORM
    record_input_type = GridoptimizationInput
    record_multi_input_type = GridoptimizationMultiInput

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseRecordSocket.__init__(self, root_socket)
        self._logger = logging.getLogger(__name__)

    def initialize_service(self, session: Session, service_orm: ServiceQueueORM) -> None:
        go_orm: GridoptimizationRecordORM = service_orm.record
        specification = GridoptimizationSpecification(**go_orm.specification.model_dict())
        keywords = specification.keywords

        # Build constraint template
        constraint_template = []
        for scan in keywords.scans:
            s = {"type": scan.type, "indices": scan.indices}
            constraint_template.append(s)

        constraint_template_str = json.dumps(constraint_template)
        dimensions = tuple(len(x.steps) for x in keywords.scans)

        if keywords.preoptimization:
            iteration = -2
        else:
            iteration = 0

        output = (
            "Created gridoptimization\n"
            f"dimensions: {dimensions}\n"
            f"preoptimization: {keywords.preoptimization}\n"
            f"starting iteration: {iteration}\n"
        )

        append_output(session, go_orm, OutputTypeEnum.stdout, output)

        service_state = GridoptimizationServiceState(
            iteration=iteration,
            complete=[],
            dimensions=dimensions,
            constraint_template=constraint_template_str,
        )

        service_orm.service_state = service_state.dict()
        sqlalchemy.orm.attributes.flag_modified(service_orm, "service_state")

    def iterate_service(
        self,
        session: Session,
        service_orm: ServiceQueueORM,
    ) -> bool:
        go_orm: GridoptimizationRecordORM = service_orm.record

        # Always update with the current provenance
        go_orm.compute_history[-1].provenance = {
            "creator": "qcfractal",
            "version": qcfractal_version,
            "routine": "qcfractal.services.gridoptimization",
        }

        # Load the state from the service_state column
        service_state = GridoptimizationServiceState(**service_orm.service_state)

        # Maps key to molecule
        next_tasks = {}

        # Special preoptimization iterations
        if service_state.iteration == -2:
            next_tasks["preoptimization"] = go_orm.initial_molecule.to_model(Molecule)
            service_state.iteration = -1
            output = "Starting preoptimization"

        elif service_state.iteration == -1:
            complete_deps = service_orm.dependencies

            if len(complete_deps) != 1:
                raise RuntimeError(f"Expected one complete task for preoptimization, but got {len(complete_deps)}")

            starting_molecule = complete_deps[0].record.final_molecule.to_model(Molecule)

            # Assign the true starting molecule and grid to the grid optimization record
            go_orm.starting_molecule_id = complete_deps[0].record.final_molecule_id
            go_orm.starting_grid = calculate_starting_grid(go_orm.specification.keywords["scans"], starting_molecule)

            opt_key = serialize_key(go_orm.starting_grid)
            next_tasks[opt_key] = starting_molecule

            # Skips the normal 0th iteration
            service_state.iteration = 1

            output = "Found finished preoptimization. Starting normal iterations"

        # Special start iteration
        elif service_state.iteration == 0:
            # We set starting_molecule to initial_molecule
            go_orm.starting_molecule_id = go_orm.initial_molecule_id
            starting_molecule = go_orm.initial_molecule.to_model(Molecule)

            go_orm.starting_grid = calculate_starting_grid(go_orm.specification.keywords["scans"], starting_molecule)

            opt_key = serialize_key(go_orm.starting_grid)
            next_tasks[opt_key] = starting_molecule

            service_state.iteration = 1

            output = "Starting first iterations"

        else:
            # Obtain complete tasks and figure out future tasks
            complete_deps = service_orm.dependencies

            # Maps keys to Molecule (for the next iteration)
            molecule_map = {}

            for dep in complete_deps:
                key = dep.extras["key"]
                molecule_map[key] = dep.record.final_molecule.to_model(Molecule)

            # Build out the new set of seeds
            complete_seeds = set(deserialize_key(dep.extras["key"]) for dep in complete_deps)

            # Store what we have already completed
            service_state.complete = list(set(service_state.complete) | complete_seeds)

            # Compute new points
            new_points_list = expand_ndimensional_grid(service_state.dimensions, complete_seeds, service_state.complete)

            for new_points in new_points_list:
                old = serialize_key(new_points[0])
                new = serialize_key(new_points[1])

                next_tasks[new] = molecule_map[old]

            output = f"Found {len(complete_deps)} optimizations:\n"
            for dep in complete_deps:
                output += f"    {dep.extras['key']}\n"

        if len(next_tasks) > 0:
            # Submit the new optimizations
            self._submit_optimizations(session, service_state, service_orm, next_tasks)

            output += f"Submitted {len(service_orm.dependencies)} new optimizations"
        else:
            output += "Grid optimization finished successfully!"

        append_output(session, go_orm, OutputTypeEnum.stdout, output)

        # Set the new service state. We must then mark it as modified
        # so that SQLAlchemy can pick up changes. This is because SQLAlchemy
        # cannot track mutations in nested dicts
        service_orm.service_state = service_state.dict()
        sqlalchemy.orm.attributes.flag_modified(service_orm, "service_state")

        # Return True to indicate that this service has successfully completed
        return len(next_tasks) == 0

    def _submit_optimizations(
        self,
        session: Session,
        service_state: GridoptimizationServiceState,
        service_orm: ServiceQueueORM,
        task_dict: Dict[str, Molecule],
    ):
        """
        Submit the next batch of optimizations for a gridoptimization service
        """

        go_orm: GridoptimizationRecordORM = service_orm.record

        # delete all existing entries in the dependency list
        service_orm.dependencies = []

        # Create an optimization input based on the new geometry and the optimization template
        opt_spec = go_orm.specification.optimization_specification.model_dict()

        # Convert to an input
        opt_spec = OptimizationSpecification(**opt_spec).dict()

        # Load the starting molecule (for absolute constraints)
        starting_molecule = None
        if go_orm.starting_molecule is not None:
            starting_molecule = go_orm.starting_molecule.to_model(Molecule)

        for key, molecule in task_dict.items():
            # Make a deep copy to prevent modifying the original ORM
            opt_spec2 = copy.deepcopy(opt_spec)

            if key == "preoptimization":
                if starting_molecule is not None:
                    raise RuntimeError("Developer error - starting molecule set when it shouldn't be!")
                # Submit the new optimization with no constraints
                meta, opt_ids = self.root_socket.records.optimization.add(
                    [molecule],
                    OptimizationSpecification(**opt_spec2),
                    service_orm.compute_tag,
                    service_orm.compute_priority,
                    go_orm.creator_user_id,
                    service_orm.find_existing,
                    session=session,
                )

            else:
                if starting_molecule is None:
                    raise RuntimeError("Developer error - starting molecule not set when it should be!")

                # Construct constraints
                constraints = json.loads(service_state.constraint_template)

                scan_indices = deserialize_key(key)

                for con_num, scan in enumerate(go_orm.specification.keywords["scans"]):
                    idx = scan_indices[con_num]
                    if scan["step_type"] == "absolute":
                        constraints[con_num]["value"] = scan["steps"][idx]
                    else:
                        # Measure absolute constraints from the starting molecule
                        constraints[con_num]["value"] = scan["steps"][idx] + starting_molecule.measure(scan["indices"])

                # update the constraints
                opt_spec2["keywords"].setdefault("constraints", {})
                opt_spec2["keywords"]["constraints"].setdefault("set", [])
                opt_spec2["keywords"]["constraints"]["set"].extend(constraints)

                # Submit the new optimization
                meta, opt_ids = self.root_socket.records.optimization.add(
                    [molecule],
                    OptimizationSpecification(**opt_spec2),
                    service_orm.compute_tag,
                    service_orm.compute_priority,
                    go_orm.creator_user_id,
                    service_orm.find_existing,
                    session=session,
                )

            if not meta.success:
                raise RuntimeError("Error adding optimization - likely a developer error: " + meta.error_string)

            svc_dep = ServiceDependencyORM(
                record_id=opt_ids[0],
                extras={"key": key},
            )

            # Update the association table
            opt_assoc = GridoptimizationOptimizationORM(
                optimization_id=opt_ids[0], gridoptimization_id=service_orm.record_id, key=key
            )

            service_orm.dependencies.append(svc_dep)
            go_orm.optimizations.append(opt_assoc)

    def insert_full_qcportal_records_v1(
        self,
        session: Session,
        records: Sequence[GridoptimizationRecord],
        creator_user_id: Optional[int],
    ) -> List[GridoptimizationRecordORM]:
        ret = []

        initial_mols = []
        starting_mols = []
        go_specs = []

        for record in records:
            initial_mols.append(record.initial_molecule)

            # Starting molecule should always be set for a completed gridoptimization
            starting_mols.append(record.starting_molecule)
            go_specs.append(record.specification)

        meta, spec_ids = self.root_socket.records.gridoptimization.add_specifications(go_specs, session=session)
        if not meta.success:
            raise RuntimeError(
                "Aborted gridoptimization insertion - could not add specifications: " + meta.error_string
            )

        meta, initial_mol_ids = self.root_socket.molecules.add(initial_mols, session=session)
        if not meta.success:
            raise RuntimeError(
                "Aborted gridoptimization insertion - could not add initial molecules: " + meta.error_string
            )

        meta, starting_mol_ids = self.root_socket.molecules.add(starting_mols, session=session)
        if not meta.success:
            raise RuntimeError(
                "Aborted gridoptimization insertion - could not add starting molecules: " + meta.error_string
            )

        for record, initial_mol_id, starting_mol_id, spec_id in zip(
            records, initial_mol_ids, starting_mol_ids, spec_ids
        ):
            record_orm = GridoptimizationRecordORM(
                specification_id=spec_id,
                initial_molecule_id=initial_mol_id,
                starting_molecule_id=starting_mol_id,
                starting_grid=record.starting_grid,
                status=RecordStatusEnum.complete,
            )

            if record.optimizations:
                opt_records_lst = list(record.optimization_records_.items())
                opt_records = [x[1] for x in opt_records_lst]

                opt_ids = self.root_socket.records.insert_full_qcportal_records(session, opt_records, creator_user_id)
                record_orm.optimizations = [
                    GridoptimizationOptimizationORM(
                        optimization_id=oid,
                        key=k,
                    )
                    for (k, _), oid in zip(opt_records_lst, opt_ids)
                ]

            ret.append(record_orm)

        return ret

    def add_specifications(
        self, go_specs: Sequence[GridoptimizationSpecification], *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, List[int]]:
        """
        Adds specifications for gridoptimization services to the database, returning their IDs.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Specification IDs are returned in the same order as the input specifications

        Parameters
        ----------
        go_specs
            Sequence of specifications to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the IDs of the specifications.
        """

        to_add = []

        for go_spec in go_specs:
            go_kw_dict = go_spec.keywords.dict()

            go_spec_dict = {"program": go_spec.program, "keywords": go_kw_dict, "protocols": {}}
            go_spec_hash = hash_dict(go_spec_dict)

            go_spec_orm = GridoptimizationSpecificationORM(
                program=go_spec.program,
                keywords=go_kw_dict,
                protocols=go_spec_dict["protocols"],
                specification_hash=go_spec_hash,
            )

            to_add.append(go_spec_orm)

        with self.root_socket.optional_session(session, False) as session:

            opt_specs = [x.optimization_specification for x in go_specs]
            meta, opt_spec_ids = self.root_socket.records.optimization.add_specifications(opt_specs, session=session)

            if not meta.success:
                return (
                    InsertMetadata(
                        error_description="Unable to add optimization specifications: " + meta.error_string,
                    ),
                    [],
                )

            assert len(opt_spec_ids) == len(go_specs)
            for go_spec_orm, opt_spec_id in zip(to_add, opt_spec_ids):
                go_spec_orm.optimization_specification_id = opt_spec_id

            meta, ids = insert_general(
                session,
                to_add,
                (
                    GridoptimizationSpecificationORM.specification_hash,
                    GridoptimizationSpecificationORM.optimization_specification_id,
                ),
                (GridoptimizationSpecificationORM.id,),
                use_unique=True,
            )

            return meta, [x[0] for x in ids]

    def add_specification(
        self, go_spec: GridoptimizationSpecification, *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, Optional[int]]:
        """
        Adds a specification for a gridoptimization service to the database, returning its id.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Parameters
        ----------
        go_spec
            Specification to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the id of the specification.
        """

        meta, ids = self.add_specifications([go_spec], session=session)

        if not ids:
            return meta, None

        return meta, ids[0]

    def get(
        self,
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        options = []

        if include:
            if is_included("initial_molecule", include, exclude, False) or is_included(
                "molecules", include, exclude, False
            ):
                options.append(joinedload(GridoptimizationRecordORM.initial_molecule))
            if is_included("starting_molecule", include, exclude, False) or is_included(
                "molecules", include, exclude, False
            ):
                options.append(joinedload(GridoptimizationRecordORM.starting_molecule))
            if is_included("optimizations", include, exclude, False):
                options.append(selectinload(GridoptimizationRecordORM.optimizations))

        with self.root_socket.optional_session(session, True) as session:
            return self.root_socket.records.get_base(
                orm_type=self.record_orm,
                record_ids=record_ids,
                include=include,
                exclude=exclude,
                missing_ok=missing_ok,
                additional_options=options,
                session=session,
            )

    def query(
        self,
        query_data: GridoptimizationQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        and_query = []
        need_spspec_join = False
        need_optspec_join = False

        if query_data.qc_program is not None:
            and_query.append(QCSpecificationORM.program.in_(query_data.qc_program))
            need_spspec_join = True
        if query_data.qc_method is not None:
            and_query.append(QCSpecificationORM.method.in_(query_data.qc_method))
            need_spspec_join = True
        if query_data.qc_basis is not None:
            and_query.append(QCSpecificationORM.basis.in_(query_data.qc_basis))
            need_spspec_join = True
        if query_data.optimization_program is not None:
            and_query.append(OptimizationSpecificationORM.program.in_(query_data.optimization_program))
            need_optspec_join = True
        if query_data.initial_molecule_id is not None:
            and_query.append(GridoptimizationRecordORM.initial_molecule_id.in_(query_data.initial_molecule_id))

        stmt = select(GridoptimizationRecordORM.id)

        if need_optspec_join or need_spspec_join:
            stmt = stmt.join(GridoptimizationRecordORM.specification)
            stmt = stmt.join(GridoptimizationSpecificationORM.optimization_specification)

        if need_spspec_join:
            stmt = stmt.join(OptimizationSpecificationORM.qc_specification)

        stmt = stmt.where(*and_query)

        return self.root_socket.records.query_base(
            stmt=stmt,
            orm_type=GridoptimizationRecordORM,
            query_data=query_data,
            session=session,
        )

    def add_internal(
        self,
        initial_molecule_ids: Sequence[int],
        go_spec_id: int,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Internal function for adding new gridoptimization computations

        This function expects that the molecules and specification are already added to the
        database and that the ids are known.

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_molecule_ids
            IDs of the initial molecules to start the gridoptimizations. One record will be added per molecule.
        go_spec_id
            ID of the specification
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user_id
            ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        compute_tag = compute_tag.lower()

        with self.root_socket.optional_session(session, False) as session:
            # Lock for the entire transaction
            session.execute(select(func.pg_advisory_xact_lock(gridoptimization_insert_lock_id))).scalar()

            all_orm = []
            for mid in initial_molecule_ids:
                go_orm = GridoptimizationRecordORM(
                    is_service=True,
                    specification_id=go_spec_id,
                    initial_molecule_id=mid,
                    status=RecordStatusEnum.waiting,
                    creator_user_id=creator_user_id,
                )

                self.create_service(go_orm, compute_tag, compute_priority, find_existing)
                all_orm.append(go_orm)

            if find_existing:
                meta, ids = insert_general(
                    session,
                    all_orm,
                    (GridoptimizationRecordORM.specification_id, GridoptimizationRecordORM.initial_molecule_id),
                    (GridoptimizationRecordORM.id,),
                    use_unique=False,
                    lock_id=gridoptimization_insert_lock_id,
                )
                return meta, [x[0] for x in ids]
            else:
                session.add_all(all_orm)
                session.flush()
                meta = InsertMetadata(inserted_idx=list(range(len(all_orm))))

                return meta, [x.id for x in all_orm]

    def add(
        self,
        initial_molecules: Sequence[Union[int, Molecule]],
        go_spec: GridoptimizationSpecification,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Adds new gridoptimization calculations

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        If session is specified, changes are not committed to to the database, but the session is flushed.

        Parameters
        ----------
        initial_molecules
            Molecules to compute using the specification
        go_spec
            Specification for the calculations
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user
            Name or ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        with self.root_socket.optional_session(session, False) as session:
            creator_user_id = self.root_socket.users.get_optional_user_id(creator_user, session=session)

            # First, add the specification
            spec_meta, spec_id = self.add_specification(go_spec, session=session)
            if not spec_meta.success:
                return (
                    InsertMetadata(
                        error_description="Aborted - could not add specification: " + spec_meta.error_string
                    ),
                    [],
                )

            # Now the molecules
            mol_meta, init_mol_ids = self.root_socket.molecules.add_mixed(initial_molecules, session=session)
            if not mol_meta.success:
                return (
                    InsertMetadata(error_description="Aborted - could not add all molecules: " + mol_meta.error_string),
                    [],
                )

            return self.add_internal(
                init_mol_ids,
                spec_id,
                compute_tag,
                compute_priority,
                creator_user_id,
                find_existing,
                session=session,
            )

    def add_from_input(
        self,
        record_input: GridoptimizationInput,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertCountsMetadata, int]:

        assert isinstance(record_input, GridoptimizationInput)

        meta, ids = self.add(
            [record_input.initial_molecule],
            record_input.specification,
            compute_tag,
            compute_priority,
            creator_user,
            find_existing,
        )

        return InsertCountsMetadata.from_insert_metadata(meta), ids[0]

    ####################################################
    # Some stuff to be retrieved for gridoptimizations
    ####################################################

    def get_optimizations(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        options = [lazyload("*"), defer("*"), joinedload(GridoptimizationRecordORM.optimizations).options(undefer("*"))]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(GridoptimizationRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.model_dict() for x in rec.optimizations]



from typing import List

from flask import current_app, g

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.exceptions import LimitExceededError
from qcportal.optimization import (
    OptimizationDatasetSpecification,
    OptimizationDatasetNewEntry,
    OptimizationAddBody,
    OptimizationQueryFilters,
)
from qcportal.utils import calculate_limit


#####################
# Record
#####################


@api_v1.route("/records/optimization/bulkCreate", methods=["POST"])
@wrap_global_route("records", "add")
def add_optimization_records_v1(body_data: OptimizationAddBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.add_records
    if len(body_data.initial_molecules) > limit:
        raise LimitExceededError(
            f"Cannot add {len(body_data.initial_molecules)} optimization records - limit is {limit}"
        )

    return storage_socket.records.optimization.add(
        initial_molecules=body_data.initial_molecules,
        opt_spec=body_data.specification,
        compute_tag=body_data.compute_tag,
        compute_priority=body_data.compute_priority,
        creator_user=g.username,
        find_existing=body_data.find_existing,
    )


@api_v1.route("/records/optimization/<int:record_id>/trajectory", methods=["GET"])
@wrap_global_route("records", "read")
def get_optimization_trajectory_ids_v1(record_id: int):
    return storage_socket.records.optimization.get_trajectory_ids(record_id)


@api_v1.route("/records/optimization/query", methods=["POST"])
@wrap_global_route("records", "read")
def query_optimization_v1(body_data: OptimizationQueryFilters):
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_records
    body_data.limit = calculate_limit(max_limit, body_data.limit)

    return storage_socket.records.optimization.query(body_data)


#####################
# Dataset
#####################


@api_v1.route("/datasets/optimization/<int:dataset_id>/specifications", methods=["POST"])
@wrap_global_route("datasets", "modify")
def add_optimization_dataset_specifications_v1(dataset_id: int, body_data: List[OptimizationDatasetSpecification]):
    return storage_socket.datasets.optimization.add_specifications(dataset_id, body_data)


@api_v1.route("/datasets/optimization/<int:dataset_id>/entries/bulkCreate", methods=["POST"])
@wrap_global_route("datasets", "modify")
def add_optimization_dataset_entries_v1(dataset_id: int, body_data: List[OptimizationDatasetNewEntry]):
    return storage_socket.datasets.optimization.add_entries(dataset_id, new_entries=body_data)


@api_v1.route("/datasets/optimization/<int:dataset_id>/background_add_entries", methods=["POST"])
@wrap_global_route("datasets", "modify")
def background_add_optimization_dataset_entries_v1(dataset_id: int, body_data: List[OptimizationDatasetNewEntry]):
    return storage_socket.datasets.optimization.background_add_entries(dataset_id, new_entries=body_data)



from __future__ import annotations

import time
from typing import Optional

import pytest
import pytz

from qcfractal.components.testing_helpers import populate_records_status
from qcportal import PortalClient
from qcportal.molecules import Molecule
from qcportal.record_models import RecordStatusEnum


@pytest.fixture(scope="module")
def queryable_records_client(session_snowflake):
    # First populate all the statuses
    populate_records_status(session_snowflake.get_storage_socket())

    # Now a bunch of records
    client = session_snowflake.client()

    elements = ["h", "he", "li", "be", "b", "c", "n", "o", "f", "ne"]

    all_mols = []
    for el in elements:
        m = Molecule(
            symbols=[el],
            geometry=[0, 0, 0],
            identifiers={
                "smiles": f"madeupsmiles_{el}",
                "inchikey": f"madeupinchi_{el}",
            },
        )
        all_mols.append(m)

    all_ids = []
    for prog in ["prog1", "prog2"]:
        for driver in ["energy", "properties"]:
            for method in ["hf", "b3lyp"]:
                for basis in ["sto-3g", "def2-tzvp"]:
                    for kw in [{"maxiter": 100}, None]:
                        meta, ids = client.add_singlepoints(all_mols, prog, driver, method, basis, kw)
                        assert meta.success
                        all_ids.extend(ids)

        # prevents spurious test errors. On fast machines,
        # records can be created too close together
        # (we test at this boundary)
        time.sleep(0.05)

    assert len(all_ids) == 320
    yield client
    session_snowflake.reset()


def test_record_client_query(queryable_records_client: PortalClient):
    query_res = queryable_records_client.query_records(record_type="singlepoint")
    all_records = list(query_res)
    assert len(all_records) == 325

    # Get some ids from the last query
    ids = [x.id for x in all_records]
    query_res = queryable_records_client.query_records(record_id=ids[5:15])
    query_res_l = list(query_res)
    assert len(query_res_l) == 10

    # Created/modified before/after
    sorted_records = sorted(all_records, key=lambda x: x.created_on)
    query_res = queryable_records_client.query_records(
        record_type="singlepoint", created_before=sorted_records[164].created_on
    )
    query_res_l = list(query_res)
    assert len(query_res_l) == 165

    query_res = queryable_records_client.query_records(
        record_type="singlepoint", created_after=sorted_records[165].created_on
    )
    query_res_l = list(query_res)
    assert len(query_res_l) == 160

    sorted_records = sorted(all_records, key=lambda x: x.modified_on)
    query_res = queryable_records_client.query_records(
        record_type="singlepoint", modified_before=sorted_records[164].modified_on
    )
    query_res_l = list(query_res)
    assert len(query_res_l) == 165

    query_res = queryable_records_client.query_records(
        record_type="singlepoint", modified_after=sorted_records[165].modified_on
    )
    query_res_l = list(query_res)
    assert len(query_res_l) == 160

    # find the manager used to compute these
    manager_res = queryable_records_client.query_managers()
    manager = list(manager_res)[0]
    query_res = queryable_records_client.query_records(manager_name=manager.name)
    query_res_l = list(query_res)
    assert len(query_res_l) == 4

    # Querying based on status
    query_res = queryable_records_client.query_records(status=[RecordStatusEnum.error])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = queryable_records_client.query_records(status=RecordStatusEnum.cancelled)
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = queryable_records_client.query_records(status=[RecordStatusEnum.error, RecordStatusEnum.deleted])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # Some combinations
    query_res = queryable_records_client.query_records(record_type=["singlepoint"], status=[RecordStatusEnum.waiting])
    query_res_l = list(query_res)
    assert len(query_res_l) == 320

    query_res = queryable_records_client.query_records(record_type=["optimization"], status=[RecordStatusEnum.error])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    # Including fields
    query_res = queryable_records_client.query_records(status=RecordStatusEnum.error)
    recs = list(query_res)
    assert len(recs) == 1
    assert recs[0].task_ is None
    assert recs[0].compute_history[0].outputs_ is None

    query_res = queryable_records_client.query_records(
        status=RecordStatusEnum.error, include=["compute_history", "task"]
    )
    recs = list(query_res)
    assert recs[0].task_ is not None
    assert recs[0].compute_history_ is not None


@pytest.mark.parametrize("timezone", [None, "UTC", "Asia/Singapore"])
def test_record_client_query_timezones(queryable_records_client: PortalClient, timezone: Optional[str]):
    tzinfo = pytz.timezone(timezone) if timezone is not None else None
    query_res = queryable_records_client.query_records(record_type="singlepoint")
    all_records = list(query_res)
    assert len(all_records) == 325

    sorted_records = sorted(all_records, key=lambda x: x.created_on)
    created_before = sorted_records[164].created_on.astimezone(tzinfo)
    # print("CREATED BEFORE", created_before)
    query_res = queryable_records_client.query_records(record_type="singlepoint", created_before=created_before)
    query_res_l = list(query_res)
    assert len(query_res_l) == 165

    created_after = sorted_records[165].created_on.astimezone(tzinfo)
    # print("CREATED AFTER", created_after)
    query_res = queryable_records_client.query_records(record_type="singlepoint", created_after=created_after)
    query_res_l = list(query_res)
    assert len(query_res_l) == 160

    sorted_records = sorted(all_records, key=lambda x: x.modified_on)
    modified_before = sorted_records[164].modified_on.astimezone(tzinfo)
    # print("MODIFIED BEFORE", modified_before)
    query_res = queryable_records_client.query_records(record_type="singlepoint", modified_before=modified_before)
    query_res_l = list(query_res)
    assert len(query_res_l) == 165

    modified_after = sorted_records[165].modified_on.astimezone(tzinfo)
    # print("MODIFIED AFTER", modified_after)
    query_res = queryable_records_client.query_records(record_type="singlepoint", modified_after=modified_after)
    query_res_l = list(query_res)
    assert len(query_res_l) == 160


def test_record_client_query_empty_iter(queryable_records_client: PortalClient):
    # Empty query
    query_res = queryable_records_client.query_records()
    assert len(query_res._current_batch) < queryable_records_client.api_limits["get_records"]

    all_recs = list(query_res)
    assert len(all_recs) == 327


def test_record_client_query_limit(queryable_records_client: PortalClient):
    query_res = queryable_records_client.query_records(record_type="singlepoint", limit=100)
    all_recs = list(query_res)
    assert len(all_recs) == 100

    query_res = queryable_records_client.query_records(record_type="singlepoint", limit=50)
    all_recs = list(query_res)
    assert len(all_recs) == 50



from qcarchivetesting import load_hash_test_data
from qcfractal.components.optimization.record_db_models import OptimizationSpecificationORM
from qcfractal.components.testing_fixtures import spec_test_runner
from qcfractal.db_socket import SQLAlchemySocket
from qcportal.optimization import OptimizationSpecification, OptimizationProtocols
from qcportal.singlepoint import QCSpecification, SinglepointDriver, SinglepointProtocols

common_qc_spec = QCSpecification(
    program="prog1",
    driver=SinglepointDriver.energy,
    method="b3lyp",
    basis="6-31G*",
    keywords={"k": 1.0e-8},
    protocols=SinglepointProtocols(),
)


def test_optimization_hash_canaries(storage_socket: SQLAlchemySocket):
    # Test data is hash : spec dict
    test_data = load_hash_test_data("optimization_specification_tests")

    # Hashes are independent of qc spec
    spec_map = [(k, OptimizationSpecification(qc_specification=common_qc_spec, **v)) for k, v in test_data.items()]

    specs = [x[1] for x in spec_map]
    meta, ids = storage_socket.records.optimization.add_specifications(specs)
    assert meta.success
    assert len(ids) == len(specs)
    assert meta.n_existing == 0

    with storage_socket.session_scope() as session:
        for spec_id, (spec_hash, _) in zip(ids, spec_map):
            spec_orm = session.get(OptimizationSpecificationORM, spec_id)
            assert spec_orm.specification_hash == spec_hash


def test_optimization_socket_add_specification_same_0(spec_test_runner):
    spec1 = OptimizationSpecification(
        program="optprog1",
        keywords={"k": "value"},
        protocols=OptimizationProtocols(),
        qc_specification=common_qc_spec,
    )

    spec_test_runner("optimization", spec1, spec1, True)


def test_optimization_socket_add_specification_same_1(spec_test_runner):
    # Test case sensitivity
    spec1 = OptimizationSpecification(
        program="optprog1",
        keywords={"k": "value"},
        protocols=OptimizationProtocols(),
        qc_specification=QCSpecification(
            program="prog1",
            driver=SinglepointDriver.energy,
            method="b3lyp",
            basis="6-31G*",
            keywords={"k": 1.0e-8},
            protocols=SinglepointProtocols(),
        ),
    )

    spec2 = OptimizationSpecification(
        program="optPRog1",
        keywords={"k": "value"},
        protocols=OptimizationProtocols(),
        qc_specification=QCSpecification(
            program="PROG1",
            driver=SinglepointDriver.energy,
            method="b3LYP",
            basis="6-31g*",
            keywords={"k": 1.0e-8},
            protocols=SinglepointProtocols(),
        ),
    )

    spec_test_runner("optimization", spec1, spec2, True)


def test_optimization_socket_add_specification_same_2(spec_test_runner):
    # Test keywords defaults
    spec1 = OptimizationSpecification(
        program="optprog1", keywords={}, protocols=OptimizationProtocols(), qc_specification=common_qc_spec
    )

    spec2 = OptimizationSpecification(
        program="optprog1", protocols=OptimizationProtocols(), qc_specification=common_qc_spec
    )

    spec_test_runner("optimization", spec1, spec2, True)


def test_optimization_socket_add_specification_same_3(spec_test_runner):
    # Test protocols defaults
    spec1 = OptimizationSpecification(program="optprog1", keywords={}, qc_specification=common_qc_spec)

    spec2 = OptimizationSpecification(
        program="optprog1", keywords={}, protocols=OptimizationProtocols(), qc_specification=common_qc_spec
    )

    spec_test_runner("optimization", spec1, spec2, True)


def test_optimization_socket_add_specification_same_4(spec_test_runner):
    # Test protocols defaults (due to exclude_defaults)
    spec1 = OptimizationSpecification(program="optprog1", keywords={}, qc_specification=common_qc_spec)

    spec2 = OptimizationSpecification(
        program="optprog1",
        keywords={},
        protocols=OptimizationProtocols(trajectory="all"),
        qc_specification=common_qc_spec,
    )

    spec_test_runner("optimization", spec1, spec2, True)


def test_optimization_socket_add_specification_same_5(spec_test_runner):
    # Test keyword ordering
    spec1 = OptimizationSpecification(
        program="optprog1", keywords={"a": 10, "b": "str"}, qc_specification=common_qc_spec
    )
    spec2 = OptimizationSpecification(
        program="optprog1", keywords={"b": "str", "a": 10}, qc_specification=common_qc_spec
    )

    spec_test_runner("optimization", spec1, spec2, True)


def test_optimization_socket_add_specification_diff_1(spec_test_runner):
    # Test different keywords
    spec1 = OptimizationSpecification(
        program="optprog1",
        keywords={"k": "value"},
        protocols=OptimizationProtocols(),
        qc_specification=common_qc_spec,
    )

    spec2 = spec1.copy(update={"keywords": {"k2": "value2"}})

    spec_test_runner("optimization", spec1, spec2, False)


def test_optimization_socket_add_specification_diff_2(spec_test_runner):
    # Test different keywords
    spec1 = OptimizationSpecification(
        program="optprog1",
        protocols=OptimizationProtocols(),
        qc_specification=common_qc_spec,
    )

    spec2 = spec1.copy(update={"keywords": {"k2": "value2"}})

    spec_test_runner("optimization", spec1, spec2, False)


def test_optimization_socket_add_specification_diff_3(spec_test_runner):
    # Test different keywords
    spec1 = OptimizationSpecification(
        program="optprog1",
        keywords={"k": 1.0e-8},
        protocols=OptimizationProtocols(),
        qc_specification=common_qc_spec,
    )

    spec2 = spec1.copy(update={"keywords": {"k": 1.0e-9}})

    spec_test_runner("optimization", spec1, spec2, False)


def test_optimization_socket_add_specification_diff_4(spec_test_runner):
    # Test qc spec changes
    spec1 = OptimizationSpecification(
        program="optprog1",
        keywords={"k": 1.0e-8},
        protocols=OptimizationProtocols(),
        qc_specification=common_qc_spec,
    )

    qc_spec = common_qc_spec.copy(update={"keywords": {"a": 1}})
    spec2 = spec1.copy(update={"qc_specification": qc_spec})

    spec_test_runner("optimization", spec1, spec2, False)


def test_optimization_socket_add_specification_diff_5(spec_test_runner):
    # Test qc spec changes
    spec1 = OptimizationSpecification(
        program="optprog1",
        keywords={"k": 1.0e-8},
        protocols=OptimizationProtocols(),
        qc_specification=common_qc_spec,
    )

    qc_spec = common_qc_spec.copy(update={"protocols": SinglepointProtocols(stdout=False)})
    spec2 = spec1.copy(update={"qc_specification": qc_spec})

    spec_test_runner("optimization", spec1, spec2, False)


def test_optimization_socket_add_specification_diff_6(spec_test_runner):
    # Test different protocols
    spec1 = OptimizationSpecification(
        program="optprog1",
        keywords={"k": "value"},
        protocols=OptimizationProtocols(),
        qc_specification=common_qc_spec,
    )

    spec2 = OptimizationSpecification(
        program="optprog1",
        keywords={"k": "value"},
        protocols=OptimizationProtocols(trajectory="initial_and_final"),
        qc_specification=common_qc_spec,
    )

    spec_test_runner("optimization", spec1, spec2, False)



import copy
import os

from qcfractal.config import FractalConfig

_base_config = {
    "api": {
        "secret_key": "abc1234def456",
        "jwt_secret_key": "abc123def456",
    },
    "database": {"username": "qcfractal", "password": "abc123def456"},
}


def test_config_durations_plain(tmp_path):
    base_folder = str(tmp_path)

    base_config = copy.deepcopy(_base_config)
    base_config["service_frequency"] = 3600
    base_config["heartbeat_frequency"] = 30
    base_config["access_log_keep"] = 31
    base_config["internal_job_keep"] = 7
    base_config["api"]["jwt_access_token_expires"] = 7450
    base_config["api"]["jwt_refresh_token_expires"] = 637277
    cfg = FractalConfig(base_folder=base_folder, **base_config)

    assert cfg.service_frequency == 3600
    assert cfg.heartbeat_frequency == 30
    assert cfg.access_log_keep == 2678400  # interpreted as days
    assert cfg.internal_job_keep == 604800
    assert cfg.api.jwt_access_token_expires == 7450
    assert cfg.api.jwt_refresh_token_expires == 637277


def test_config_durations_str(tmp_path):
    base_folder = str(tmp_path)

    base_config = copy.deepcopy(_base_config)
    base_config["service_frequency"] = "1h"
    base_config["heartbeat_frequency"] = "30s"
    base_config["access_log_keep"] = "1d4h2s"
    base_config["internal_job_keep"] = "1d4h7s"
    base_config["api"]["jwt_access_token_expires"] = "2h4m10s"
    base_config["api"]["jwt_refresh_token_expires"] = "7d9h77s"
    cfg = FractalConfig(base_folder=base_folder, **base_config)

    assert cfg.service_frequency == 3600
    assert cfg.heartbeat_frequency == 30
    assert cfg.access_log_keep == 100802
    assert cfg.internal_job_keep == 100807
    assert cfg.api.jwt_access_token_expires == 7450
    assert cfg.api.jwt_refresh_token_expires == 637277


def test_config_durations_dhms(tmp_path):
    base_folder = str(tmp_path)

    base_config = copy.deepcopy(_base_config)
    base_config["service_frequency"] = "1:00:00"
    base_config["heartbeat_frequency"] = "30"
    base_config["access_log_keep"] = "1:04:00:02"
    base_config["internal_job_keep"] = "1:04:00:07"
    base_config["api"]["jwt_access_token_expires"] = "2:04:10"
    base_config["api"]["jwt_refresh_token_expires"] = "7:09:00:77"
    cfg = FractalConfig(base_folder=base_folder, **base_config)

    assert cfg.service_frequency == 3600
    assert cfg.heartbeat_frequency == 30
    assert cfg.access_log_keep == 100802
    assert cfg.internal_job_keep == 100807
    assert cfg.api.jwt_access_token_expires == 7450
    assert cfg.api.jwt_refresh_token_expires == 637277


def test_config_tmpdir_create(tmp_path):
    base_config = copy.deepcopy(_base_config)
    base_config["base_folder"] = str(tmp_path)
    base_config["temporary_dir"] = str(tmp_path / "qcatmpdir")
    cfg = FractalConfig(**base_config)

    assert cfg.temporary_dir == str(tmp_path / "qcatmpdir")
    assert os.path.exists(cfg.temporary_dir)



from typing import Optional

import pytest

from qcarchivetesting import test_users
from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcportal import PortalRequestError
from qcportal.auth import UserInfo
from qcportal.exceptions import (
    InvalidUsernameError,
    InvalidPasswordError,
    AuthenticationFailure,
)
from .test_group_socket import invalid_groupnames
from .test_user_socket import invalid_usernames, invalid_passwords


def test_user_client_list(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])
    users = client.list_users()

    assert len(users) == len(test_users)

    for u in users:
        assert u.username in test_users

        tu = test_users[u.username]["info"]
        assert u.role == tu["role"]
        assert u.fullname == tu["fullname"]
        assert u.organization == tu["organization"]
        assert u.email == tu["email"]


def test_user_client_get(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    for username, uinfo in test_users.items():
        u = client.get_user(username)

        assert u.username == username
        assert u.role == uinfo["info"]["role"]

        u2 = client.get_user(u.id)
        assert u2 == u


def test_user_client_get_me(secure_snowflake: QCATestingSnowflake):
    for username, uinfo in test_users.items():
        client = secure_snowflake.client(username, uinfo["pw"])
        me = client.get_user()

        assert me.username == username
        assert me.role == uinfo["info"]["role"]


def test_user_client_add(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    uinfo = UserInfo(
        username="george",
        role="compute",
        enabled=True,
        fullname="Test user",
        email="george@example.com",
        organization="My Org",
    )

    pw = client.add_user(uinfo)

    secure_snowflake.client("george", pw)


def test_user_client_add_existing(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    uinfo = UserInfo(
        username="read_user",
        role="compute",
        enabled=True,
        fullname="Test user",
        email="george@example.com",
        organization="My Org",
    )

    with pytest.raises(PortalRequestError, match=r"already exists"):
        client.add_user(uinfo)


def test_user_client_add_badrole(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    uinfo = UserInfo(
        username="read_user",
        role="does_not_exist",
        enabled=True,
        fullname="Test user",
        email="george@example.com",
        organization="My Org",
    )

    with pytest.raises(PortalRequestError, match=r"Role.*does not exist"):
        client.add_user(uinfo)


def test_user_client_add_badgroup(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    uinfo = UserInfo(
        username="read_user_2",
        role="read",
        groups=["group_does_not_exist"],
        enabled=True,
    )

    with pytest.raises(PortalRequestError, match=r"Group.*does not exist"):
        client.add_user(uinfo)


def test_user_client_use_nonexist(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    uinfo = client.get_user()
    uinfo.__dict__["id"] = 1234  # bypass pydantic validation

    with pytest.raises(PortalRequestError, match=r"User.*not found"):
        client.get_user("no_user")
    with pytest.raises(PortalRequestError, match=r"User.*not found"):
        client.modify_user(uinfo)
    with pytest.raises(PortalRequestError, match=r"User.*not found"):
        client.change_user_password("no_user", "abcde1234")
    with pytest.raises(PortalRequestError, match=r"User.*not found"):
        client.delete_user("no_user")


def test_user_client_use_invalid_username(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    for username in invalid_usernames:
        uinfo = client.get_user()
        uinfo.__dict__["username"] = username  # bypass pydantic validation
        uinfo.__dict__["id"] = None

        with pytest.raises(InvalidUsernameError):
            client.get_user(username)
        with pytest.raises(PortalRequestError, match=r"Username"):
            client.add_user(uinfo)
        with pytest.raises(PortalRequestError, match=r"Username"):
            client.modify_user(uinfo)
        with pytest.raises(InvalidUsernameError):
            client.change_user_password(username, "abcde1234")
        with pytest.raises(InvalidUsernameError):
            client.delete_user(username)


def test_user_client_use_invalid_groupname(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    for groupname in invalid_groupnames:
        uinfo = client.get_user()
        uinfo.__dict__["groups"] = [groupname]  # bypass pydantic validation
        uinfo.__dict__["id"] = None

        with pytest.raises(PortalRequestError, match=r"Groupname"):
            client.add_user(uinfo)
        with pytest.raises(PortalRequestError, match=r"Groupname"):
            client.modify_user(uinfo)


def test_user_client_use_invalid_password(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    for password in invalid_passwords:
        uinfo = client.get_user()
        uinfo.__dict__["username"] = "new_user"  # bypass pydantic validation

        with pytest.raises(InvalidPasswordError):
            client.add_user(uinfo, password)
        with pytest.raises(InvalidPasswordError):
            client.change_user_password(None, password)


def test_user_client_delete(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    client.delete_user("read_user")

    with pytest.raises(PortalRequestError, match=r"User.*not found"):
        client.get_user("read_user")


def test_user_client_delete_self(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    with pytest.raises(PortalRequestError, match=r"Cannot delete your own user"):
        client.delete_user("admin_user")


def test_user_client_modify(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])
    uinfo = client.get_user("read_user")

    uinfo.fullname = "New Full Name"
    uinfo.organization = "New Organization"
    uinfo.email = "New Email"
    uinfo.role = "submit"
    uinfo.groups = ["group3"]
    uinfo.enabled = False

    uinfo2 = client.modify_user(uinfo)

    uinfo3 = client.get_user("read_user")
    assert uinfo2 == uinfo
    assert uinfo3 == uinfo

    assert uinfo3.fullname == "New Full Name"
    assert uinfo3.organization == "New Organization"
    assert uinfo3.email == "New Email"
    assert uinfo3.role == "submit"
    assert uinfo3.enabled is False
    assert uinfo3.groups == ["group3"]


def test_user_client_modify_self(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("read_user", test_users["read_user"]["pw"])
    uinfo = client.get_user("read_user")

    uinfo.fullname = "New Full Name"
    uinfo.organization = "New Organization"
    uinfo.email = "New Email"
    uinfo.role = "submit"
    uinfo.groups = ["group3"]
    uinfo.enabled = False

    uinfo2 = client.modify_user(uinfo)

    uinfo3 = client.get_user("read_user")
    assert uinfo2 == uinfo3

    # Only some fields changed
    assert uinfo3.fullname == "New Full Name"
    assert uinfo3.organization == "New Organization"
    assert uinfo3.email == "New Email"

    # Did not change
    assert uinfo3.role == "read"
    assert uinfo3.enabled is True
    assert uinfo3.groups == ["group2"]


def test_user_client_modify_self_me(secure_snowflake: QCATestingSnowflake):
    # Similar to above, but using the /me endpoint
    client = secure_snowflake.client("read_user", test_users["read_user"]["pw"])
    uinfo = client.get_user("read_user")

    uinfo.fullname = "New Full Name"
    uinfo.organization = "New Organization"
    uinfo.email = "New Email"
    uinfo.role = "submit"
    uinfo.groups = ["group3"]
    uinfo.enabled = False

    uinfo2 = client.make_request("patch", f"api/v1/me", UserInfo, body=uinfo)
    uinfo3 = client.get_user("read_user")
    assert uinfo2 == uinfo3

    # Only some fields changed
    assert uinfo3.fullname == "New Full Name"
    assert uinfo3.organization == "New Organization"
    assert uinfo3.email == "New Email"

    # Did not change
    assert uinfo3.role == "read"
    assert uinfo3.enabled is True
    assert uinfo3.groups == ["group2"]


def test_user_client_modify_badrole(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])
    uinfo = client.get_user("read_user")

    uinfo.fullname = "New Full Name"
    uinfo.organization = "New Organization"
    uinfo.email = "New Email"
    uinfo.role = "bad_role"
    uinfo.enabled = False

    with pytest.raises(PortalRequestError, match=r"Role.*does not exist"):
        client.modify_user(uinfo)


def test_user_client_modify_badgroup(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])
    uinfo = client.get_user("read_user")

    uinfo.fullname = "New Full Name"
    uinfo.organization = "New Organization"
    uinfo.email = "New Email"
    uinfo.enabled = False
    uinfo.groups = ["does_not_exist"]

    with pytest.raises(PortalRequestError, match=r"Group.*does not exist"):
        client.modify_user(uinfo)


def test_user_client_change_password(secure_snowflake: QCATestingSnowflake):
    # First, make sure read user is denied
    with pytest.raises(AuthenticationFailure):
        secure_snowflake.client("read_user", "a_new_password1234")

    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    new_pw = client.change_user_password("read_user", "a_new_password1234")

    # Change password returns the same password
    assert new_pw == "a_new_password1234"

    # Now we can login
    secure_snowflake.client("read_user", "a_new_password1234")


def test_user_client_change_password_self(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("read_user", test_users["read_user"]["pw"])

    new_pw = client.change_user_password(None, "a_new_password1234")
    assert new_pw == "a_new_password1234"
    secure_snowflake.client("read_user", "a_new_password1234")

    # Reset password
    new_pw = client.change_user_password()
    secure_snowflake.client("read_user", new_pw)


def test_user_client_change_password_self_me(secure_snowflake: QCATestingSnowflake):
    # Similar to above, but using the /me endpoint
    # we don't expose this in qcportal, so manually do make_request

    client = secure_snowflake.client("read_user", test_users["read_user"]["pw"])

    new_password = "a_new_password1234"
    new_pw = client.make_request("put", f"api/v1/me/password", str, body_model=str, body=new_password)
    assert new_pw == "a_new_password1234"

    secure_snowflake.client("read_user", "a_new_password1234")

    # Reset password
    new_pw_2 = client.make_request("put", f"api/v1/me/password", str, body_model=Optional[str], body=None)
    secure_snowflake.client("read_user", new_pw_2)

    with pytest.raises(AuthenticationFailure, match="Incorrect username or password"):
        secure_snowflake.client("read_user", new_pw)


def test_user_client_get_set_preferences(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])
    user_id = client.get_user("read_user").id

    prefs = {"favorites": [4, 5, 2], "mode": "dark"}

    client.make_request("put", f"api/v1/users/{user_id}/preferences", None, body_model=dict, body=prefs)

    server_prefs = client.make_request("get", f"api/v1/users/{user_id}/preferences", dict)
    assert server_prefs == prefs

    prefs["new_pref"] = "value"

    client.make_request("put", f"api/v1/users/{user_id}/preferences", None, body_model=dict, body=prefs)

    server_prefs = client.make_request("get", f"api/v1/users/{user_id}/preferences", dict)
    assert server_prefs == prefs


def test_user_client_set_preferences_self(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("read_user", test_users["read_user"]["pw"])
    user_id = client.get_user("read_user").id

    prefs = {"favorites": [4, 5, 2], "mode": "dark"}

    client.make_request("put", f"api/v1/users/{user_id}/preferences", None, body_model=dict, body=prefs)

    server_prefs = client.make_request("get", f"api/v1/users/{user_id}/preferences", dict)
    assert server_prefs == prefs

    prefs["new_pref"] = "value"

    client.make_request("put", f"api/v1/users/{user_id}/preferences", None, body_model=dict, body=prefs)

    server_prefs = client.make_request("get", f"api/v1/users/{user_id}/preferences", dict)
    assert server_prefs == prefs


def test_user_client_set_preferences_self_me(secure_snowflake: QCATestingSnowflake):
    # Similar to above, but using the /me endpoint

    client = secure_snowflake.client("read_user", test_users["read_user"]["pw"])

    prefs = {"favorites": [4, 5, 2], "mode": "dark"}

    client.make_request("put", f"api/v1/me/preferences", None, body_model=dict, body=prefs)

    server_prefs = client.make_request("get", f"api/v1/me/preferences", dict)
    assert server_prefs == prefs

    prefs["new_pref"] = "value"

    client.make_request("put", f"api/v1/me/preferences", None, body_model=dict, body=prefs)

    server_prefs = client.make_request("get", f"api/v1/me/preferences", dict)
    assert server_prefs == prefs



"""Add postgres nulls not distinct to specs

Revision ID: 83127ad1a00e
Revises: 35bb042920a3
Create Date: 2026-02-27 20:40:36.102875

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "83127ad1a00e"
down_revision = "35bb042920a3"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint("ux_neb_specification_keys", "neb_specification", type_="unique")
    op.drop_constraint("ux_reaction_specification_keys", "reaction_specification", type_="unique")

    # We need to find duplicates due to a NULL optimization_specification_id
    # then update references and delete
    op.execute(sa.text("""
        WITH ranked AS (
            SELECT
                id,
                MIN(id) OVER (
                    PARTITION BY specification_hash, singlepoint_specification_id
                    ) AS canonical_id
            FROM neb_specification
            WHERE optimization_specification_id IS NULL
        ),
             dupes AS (
                 SELECT id AS duplicate_id, canonical_id
                 FROM ranked
                 WHERE id <> canonical_id
             )
        UPDATE neb_record r
        SET specification_id = d.canonical_id
        FROM dupes d
        WHERE r.specification_id = d.duplicate_id;
    """))

    op.execute(sa.text("""
        WITH ranked AS (
            SELECT
                id,
                MIN(id) OVER (
                    PARTITION BY specification_hash, singlepoint_specification_id
                    ) AS canonical_id
            FROM neb_specification
            WHERE optimization_specification_id IS NULL
        ),
             dupes AS (
                 SELECT id AS duplicate_id, canonical_id
                 FROM ranked
                 WHERE id <> canonical_id
             )
        UPDATE neb_dataset_specification rds
        SET specification_id = d.canonical_id
        FROM dupes d
        WHERE rds.specification_id = d.duplicate_id;
    """))

    op.execute(sa.text("""
        WITH ranked AS (
            SELECT
                id,
                MIN(id) OVER (
                    PARTITION BY specification_hash, singlepoint_specification_id
                    ) AS canonical_id
            FROM neb_specification
            WHERE optimization_specification_id IS NULL
        )
        DELETE FROM neb_specification r
            USING ranked d
        WHERE r.id = d.id
          AND d.id <> d.canonical_id;
    """))

    # Now for reactions, which can have singlepoint or optimization specification ids be null
    op.execute(sa.text("""
        WITH ranked AS (
            SELECT
                id,
                MIN(id) OVER (
                    PARTITION BY specification_hash, singlepoint_specification_id, optimization_specification_id
                    ) AS canonical_id
            FROM reaction_specification
        ),
             dupes AS (
                 SELECT id AS duplicate_id, canonical_id
                 FROM ranked
                 WHERE id <> canonical_id
             )
        UPDATE reaction_record r
        SET specification_id = d.canonical_id
        FROM dupes d
        WHERE r.specification_id = d.duplicate_id;
    """))

    op.execute(sa.text("""
        WITH ranked AS (
            SELECT
                id,
                MIN(id) OVER (
                    PARTITION BY specification_hash, singlepoint_specification_id, optimization_specification_id
                    ) AS canonical_id
            FROM reaction_specification
        ),
             dupes AS (
                 SELECT id AS duplicate_id, canonical_id
                 FROM ranked
                 WHERE id <> canonical_id
             )
        UPDATE reaction_dataset_specification rds
        SET specification_id = d.canonical_id
        FROM dupes d
        WHERE rds.specification_id = d.duplicate_id;
    """))

    op.execute(sa.text("""
        WITH ranked AS (
            SELECT
                id,
                MIN(id) OVER (
                    PARTITION BY specification_hash, singlepoint_specification_id, optimization_specification_id
                    ) AS canonical_id
            FROM reaction_specification
        )
        DELETE FROM reaction_specification r
            USING ranked d
        WHERE r.id = d.id
          AND d.id <> d.canonical_id;
    """))

    op.create_unique_constraint(
        "ux_neb_specification_keys",
        "neb_specification",
        ["specification_hash", "singlepoint_specification_id", "optimization_specification_id"],
        postgresql_nulls_not_distinct=True,
    )

    op.create_unique_constraint(
        "ux_reaction_specification_keys",
        "reaction_specification",
        ["specification_hash", "singlepoint_specification_id", "optimization_specification_id"],
        postgresql_nulls_not_distinct=True,
    )

    pass


def downgrade():
    op.drop_constraint("ux_neb_specification_keys", "neb_specification", type_="unique")
    op.drop_constraint("ux_reaction_specification_keys", "reaction_specification", type_="unique")

    op.create_unique_constraint(
        "ux_neb_specification_keys",
        "neb_specification",
        ["specification_hash", "singlepoint_specification_id", "optimization_specification_id"],
    )

    op.create_unique_constraint(
        "ux_reaction_specification_keys",
        "reaction_specification",
        ["specification_hash", "singlepoint_specification_id", "optimization_specification_id"],
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data, test_users
from qcfractal.components.reaction.record_db_models import ReactionRecordORM
from qcfractal.db_socket import SQLAlchemySocket
from qcfractal.testing_helpers import run_service
from qcportal.auth import UserInfo
from qcportal.reaction import ReactionSpecification, ReactionKeywords, compare_reaction_records
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import SinglepointProtocols, QCSpecification
from qcportal.utils import now_at_utc
from .testing_helpers import (
    compare_reaction_specs,
    test_specs,
    load_procedure_data,
    generate_task_key,
    load_record_data,
)

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal.managers import ManagerName
    from sqlalchemy.orm.session import Session


@pytest.mark.parametrize("spec", test_specs)
def test_reaction_socket_add_get(storage_socket: SQLAlchemySocket, session: Session, spec: ReactionSpecification):
    hooh = load_molecule_data("peroxide2")
    ne4 = load_molecule_data("neon_tetramer")
    water = load_molecule_data("water_dimer_minima")

    time_0 = now_at_utc()
    meta, ids = storage_socket.records.reaction.add(
        [[(1.0, hooh), (2.0, ne4)], [(3.0, hooh), (4.0, water)]], spec, "tag1", PriorityEnum.low, None, True
    )
    time_1 = now_at_utc()
    assert meta.success

    recs = [session.get(ReactionRecordORM, i) for i in ids]
    assert len(recs) == 2

    for r in recs:
        assert r.record_type == "reaction"
        assert r.status == RecordStatusEnum.waiting
        assert compare_reaction_specs(spec, r.specification.model_dict())

        # Service queue entry should exist with the proper tag and priority
        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    mol_hash_0 = set(x.molecule.identifiers["molecule_hash"] for x in recs[0].components)
    mol_hash_1 = set(x.molecule.identifiers["molecule_hash"] for x in recs[1].components)

    assert mol_hash_0 == {hooh.get_hash(), ne4.get_hash()}
    assert mol_hash_1 == {hooh.get_hash(), water.get_hash()}

    expected_coef = {hooh.get_hash(): 1.0, ne4.get_hash(): 2.0}
    db_coef = {x.molecule.identifiers["molecule_hash"]: x.coefficient for x in recs[0].components}
    assert expected_coef == db_coef

    expected_coef = {hooh.get_hash(): 3.0, water.get_hash(): 4.0}
    db_coef = {x.molecule.identifiers["molecule_hash"]: x.coefficient for x in recs[1].components}
    assert expected_coef == db_coef


def test_reaction_socket_add_same_1(storage_socket: SQLAlchemySocket):
    spec = ReactionSpecification(
        program="reaction",
        singlepoint_specification=QCSpecification(
            program="proG2",
            driver="energy",
            method="b3lyp",
            basis="6-31G*",
            keywords={"k": "v"},
            protocols=SinglepointProtocols(wavefunction="all"),
        ),
        keywords=ReactionKeywords(),
    )

    hooh = load_molecule_data("peroxide2")
    water = load_molecule_data("water_dimer_minima")

    meta, id1 = storage_socket.records.reaction.add(
        [[(2.0, water), (3.0, hooh)]], spec, "*", PriorityEnum.normal, None, True
    )
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.reaction.add(
        [[(3.0, hooh), (2.0, water)]], spec, "*", PriorityEnum.normal, None, True
    )
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


@pytest.mark.parametrize(
    "test_data_name",
    [
        "rxn_H2_psi4_b3lyp_sp",
        "rxn_H2O_psi4_b3lyp_sp",
        "rxn_H2O_psi4_mp2_opt",
        "rxn_H2O_psi4_mp2_optsp",
    ],
)
def test_reaction_socket_run(
    storage_socket: SQLAlchemySocket, session: Session, activated_manager_name: ManagerName, test_data_name: str
):
    input_spec_1, stoich_1, result_data_1 = load_procedure_data(test_data_name)

    storage_socket.users.add(UserInfo(username="submit_user", role="submit", enabled=True))

    meta_1, id_1 = storage_socket.records.reaction.add(
        [stoich_1], input_spec_1, "test_tag", PriorityEnum.low, "submit_user", True
    )
    id_1 = id_1[0]
    assert meta_1.success

    time_0 = now_at_utc()
    finished, n_singlepoints = run_service(
        storage_socket, activated_manager_name, id_1, generate_task_key, result_data_1, 100
    )
    time_1 = now_at_utc()

    assert finished is True

    rec = session.get(ReactionRecordORM, id_1)

    assert rec.status == RecordStatusEnum.complete
    assert time_0 < rec.modified_on < time_1
    assert len(rec.compute_history) == 1
    assert len(rec.compute_history[-1].outputs) == 1
    assert rec.compute_history[-1].status == RecordStatusEnum.complete
    assert time_0 < rec.compute_history[-1].modified_on < time_1
    assert rec.service is None

    desc_info = storage_socket.records.get_short_descriptions([id_1])[0]
    short_desc = desc_info["description"]
    assert desc_info["record_type"] == rec.record_type
    assert desc_info["created_on"] == rec.created_on
    assert rec.specification.program in short_desc
    if rec.specification.singlepoint_specification is not None:
        assert rec.specification.singlepoint_specification.program in short_desc
        assert rec.specification.singlepoint_specification.method in short_desc
    if rec.specification.optimization_specification is not None:
        assert rec.specification.optimization_specification.program in short_desc
        assert rec.specification.optimization_specification.qc_specification.method in short_desc

    out = rec.compute_history[-1].outputs["stdout"].get_output()
    assert "All reaction components are complete" in out

    assert rec.total_energy < 0.0


def test_reaction_socket_run_duplicate(
    storage_socket: SQLAlchemySocket,
    session: Session,
    activated_manager_name: ManagerName,
):
    input_spec_1, molecules_1, result_data_1 = load_procedure_data("rxn_H2O_psi4_mp2_optsp")

    meta_1, id_1 = storage_socket.records.reaction.add(
        [molecules_1], input_spec_1, "test_tag", PriorityEnum.low, None, True
    )
    id_1 = id_1[0]
    assert meta_1.success

    run_service(storage_socket, activated_manager_name, id_1, generate_task_key, result_data_1, 1000)

    rec_1 = session.get(ReactionRecordORM, id_1)
    assert rec_1.status == RecordStatusEnum.complete
    opt_ids_1 = [x.optimization_id for x in rec_1.components]
    sp_ids_1 = [x.singlepoint_id for x in rec_1.components]

    # Submit again, without duplicate checking
    meta_2, id_2 = storage_socket.records.reaction.add(
        [molecules_1], input_spec_1, "test_tag", PriorityEnum.low, None, False
    )
    id_2 = id_2[0]
    assert meta_2.success
    assert id_2 != id_1

    run_service(storage_socket, activated_manager_name, id_2, generate_task_key, result_data_1, 1000)

    rec_2 = session.get(ReactionRecordORM, id_2)
    assert rec_2.status == RecordStatusEnum.complete
    opt_ids_2 = [x.optimization_id for x in rec_2.components]
    sp_ids_2 = [x.singlepoint_id for x in rec_2.components]

    assert set(opt_ids_1 + sp_ids_1).isdisjoint(opt_ids_2 + sp_ids_2)


def test_reaction_socket_insert_full_qcportal_record(secure_snowflake: QCATestingSnowflake):
    test_names = [
        "rxn_H2O_psi4_b3lyp_sp",
        "rxn_H2O_psi4_mp2_opt",
        "rxn_H2O_psi4_mp2_optsp",
        "rxn_H2_psi4_b3lyp_sp",
        "rxn_error_118326390",
    ]

    storage_socket = secure_snowflake.get_storage_socket()
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    user_id = client.get_user().id

    for test_name in test_names:
        initial_record = load_record_data(test_name)
        initial_record_copy = initial_record.copy(deep=True)

        # Need a full copy of results - they can get mutated
        with storage_socket.session_scope() as session:
            ins_ids = storage_socket.records.insert_full_qcportal_records(session, [initial_record_copy], user_id)

        rec_1 = client.get_reactions(ins_ids[0], include=["**"])
        rec_1.fetch_children(include=["**"], force_fetch=True)
        compare_reaction_records(rec_1, initial_record)

        assert rec_1.creator_user == "submit_user"

        for c in rec_1.components:
            if c.singlepoint_record is not None:
                assert c.singlepoint_record.creator_user == "submit_user"
            if c.optimization_record is not None:
                assert c.optimization_record.creator_user == "submit_user"



from __future__ import annotations

import time
from datetime import timedelta
from typing import TYPE_CHECKING

import pytest

from qcfractal.components.gridoptimization.testing_helpers import submit_procedure_data as submit_go_procedure_data
from qcfractal.components.torsiondrive.testing_helpers import submit_procedure_data as submit_td_procedure_data
from qcportal.managers import ManagerName, ManagerStatusEnum
from qcportal.record_models import RecordStatusEnum
from qcportal.serverinfo.models import AccessLogQueryFilters
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake

pytestmark = pytest.mark.slow


def test_periodics_manager_heartbeats(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()

    heartbeat = snowflake._qcf_config.heartbeat_frequency
    max_missed = snowflake._qcf_config.heartbeat_max_missed

    snowflake.start_job_runner()

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "psi4": ["unknown"], "qchem": ["v3.0"]},
        compute_tags=["tag1"],
    )

    for i in range(max_missed + 1):
        time.sleep(heartbeat)
        manager = storage_socket.managers.get([mname1.fullname])

        if i < max_missed:
            assert manager[0]["status"] == ManagerStatusEnum.active
        else:
            assert manager[0]["status"] == ManagerStatusEnum.inactive


def test_periodics_service_iteration(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()

    id_1, _ = submit_td_procedure_data(storage_socket, "td_H2O2_mopac_pm6")

    service_freq = snowflake._qcf_config.service_frequency

    rec = storage_socket.records.get([id_1])
    assert rec[0]["status"] == RecordStatusEnum.waiting

    snowflake.start_job_runner()

    time.sleep(1.0)

    # added after startup
    id_2, _ = submit_go_procedure_data(storage_socket, "go_H2O2_psi4_b3lyp")

    # The first services iterated at startup
    rec = storage_socket.records.get([id_1, id_2])
    assert rec[0]["status"] == RecordStatusEnum.running
    assert rec[1]["status"] == RecordStatusEnum.waiting

    # wait for the next iteration. Then both should be running
    time.sleep(service_freq + 0.5)
    rec = storage_socket.records.get([id_1, id_2])
    assert rec[0]["status"] == RecordStatusEnum.running
    assert rec[1]["status"] == RecordStatusEnum.running


def test_periodics_delete_old_access_logs(secure_snowflake: QCATestingSnowflake):
    storage_socket = secure_snowflake.get_storage_socket()

    read_id = storage_socket.users.get("read_user")["id"]

    access1 = {
        "module": "api",
        "method": "GET",
        "full_uri": "/api/v1/datasets",
        "ip_address": "127.0.0.1",
        "user_agent": "Fake user agent",
        "request_duration": 0.24,
        "user_id": read_id,
        "request_bytes": 123,
        "response_bytes": 18273,
        "timestamp": now_at_utc() - timedelta(days=2),
    }

    access2 = {
        "module": "api",
        "method": "POST",
        "full_uri": "/api/v1/records",
        "ip_address": "127.0.0.2",
        "user_agent": "Fake user agent",
        "request_duration": 0.45,
        "user_id": read_id,
        "request_bytes": 456,
        "response_bytes": 12671,
        "timestamp": now_at_utc(),
    }

    storage_socket.serverinfo.save_access(access1)
    storage_socket.serverinfo.save_access(access2)

    accesses = storage_socket.serverinfo.query_access_log(AccessLogQueryFilters())
    n_access = len(accesses)

    secure_snowflake.start_job_runner()
    # There's a set delay at startup before we delete old logs
    time.sleep(5.0)

    # we only removed the really "old" (manually added) one
    accesses = storage_socket.serverinfo.query_access_log(AccessLogQueryFilters())
    assert len(accesses) == n_access - 1



from __future__ import annotations

from typing import TYPE_CHECKING

from qcelemental.models.results import WavefunctionProperties
from sqlalchemy import (
    Column,
    Integer,
    ForeignKey,
    String,
    Enum,
    UniqueConstraint,
    CheckConstraint,
    Index,
    LargeBinary,
    DDL,
    event,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship, deferred

from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.db_socket.base_orm import BaseORM
from qcportal.compression import CompressionEnum, decompress
from qcportal.singlepoint import SinglepointDriver

if TYPE_CHECKING:
    from typing import Dict, Optional


class WavefunctionORM(BaseORM):
    """
    Table for storing wavefunction data
    """

    __tablename__ = "wavefunction_store"

    id = Column(Integer, primary_key=True)
    record_id = Column(Integer, ForeignKey("singlepoint_record.id", ondelete="cascade"), nullable=False)

    compression_type = Column(Enum(CompressionEnum), nullable=False)
    compression_level = Column(Integer, nullable=False)
    data = deferred(Column(LargeBinary, nullable=False))

    __table_args__ = (UniqueConstraint("record_id", name="ux_wavefunction_store_record_id"),)

    _qcportal_model_excludes = ["id", "record_id", "compression_level"]

    def get_wavefunction(self) -> WavefunctionProperties:
        d = decompress(self.data, self.compression_type)
        return WavefunctionProperties(**d)


class QCSpecificationORM(BaseORM):
    """
    Table for storing the core specifications of a QC calculation
    """

    __tablename__ = "qc_specification"

    id = Column(Integer, primary_key=True)
    specification_hash = Column(String, nullable=False)

    program = Column(String(100), nullable=False)
    driver = Column(Enum(SinglepointDriver), nullable=False)
    method = Column(String(100), nullable=False)
    basis = Column(String(100), nullable=False)
    keywords = Column(JSONB, nullable=False)
    protocols = Column(JSONB, nullable=False)

    __table_args__ = (
        UniqueConstraint("specification_hash", name="ux_qc_specification_specification_hash"),
        Index("ix_qc_specification_program", "program"),
        Index("ix_qc_specification_driver", "driver"),
        Index("ix_qc_specification_method", "method"),
        Index("ix_qc_specification_basis", "basis"),
        # Enforce lowercase on some fields
        # This does not actually change the text to lowercase, but will fail to insert anything not lowercase
        # WARNING - these are not autodetected by alembic
        CheckConstraint("program = LOWER(program)", name="ck_qc_specification_program_lower"),
        CheckConstraint("method = LOWER(method)", name="ck_qc_specification_method_lower"),
        CheckConstraint("basis = LOWER(basis)", name="ck_qc_specification_basis_lower"),
    )

    _qcportal_model_excludes = ["id", "specification_hash"]

    @property
    def required_programs(self) -> Dict[str, Optional[str]]:
        return {self.program: None}

    @property
    def short_description(self) -> str:
        return f'{self.program}/{self.method or "(none)"}/{self.basis or "(none)"}'


class SinglepointRecordORM(BaseRecordORM):
    """
    Table for storing singlepoint calculations
    """

    __tablename__ = "singlepoint_record"

    id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), primary_key=True)

    specification_id = Column(Integer, ForeignKey(QCSpecificationORM.id), nullable=False)
    specification = relationship(QCSpecificationORM, lazy="selectin")

    molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=False)
    molecule = relationship(MoleculeORM)

    wavefunction = relationship(WavefunctionORM, uselist=False, cascade="all, delete-orphan", passive_deletes=True)

    __table_args__ = (
        Index("ix_singlepoint_record_molecule_id", "molecule_id"),
        Index("ix_singlepoint_record_specification_id", "specification_id"),
    )

    __mapper_args__ = {
        "polymorphic_identity": "singlepoint",
    }

    _qcportal_model_excludes = [*BaseRecordORM._qcportal_model_excludes, "specification_id"]

    @property
    def required_programs(self) -> Dict[str, Optional[str]]:
        programs = self.specification.required_programs
        programs["qcengine"] = None
        return programs

    @property
    def short_description(self) -> str:
        return f'{self.molecule.identifiers["molecular_formula"]} {self.specification.short_description}'


# Mark the storage of the wavefunction data_local column as external
event.listen(
    WavefunctionORM.__table__,
    "after_create",
    DDL("ALTER TABLE native_file ALTER COLUMN data SET STORAGE EXTERNAL").execute_if(dialect=("postgresql")),
)


# Delete base record if this record is deleted
_del_baserecord_trigger = DDL(
    """
    CREATE TRIGGER qca_singlepoint_record_delete_base_tr
    AFTER DELETE ON singlepoint_record
    FOR EACH ROW EXECUTE PROCEDURE qca_base_record_delete();
    """
)

event.listen(SinglepointRecordORM.__table__, "after_create", _del_baserecord_trigger.execute_if(dialect=("postgresql")))



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data, test_users
from qcfractal.components.torsiondrive.record_db_models import TorsiondriveRecordORM
from qcfractal.db_socket import SQLAlchemySocket
from qcfractal.testing_helpers import run_service
from qcportal.auth import UserInfo
from qcportal.optimization import OptimizationSpecification, OptimizationProtocols
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import QCSpecification, SinglepointProtocols
from qcportal.torsiondrive import TorsiondriveSpecification, TorsiondriveKeywords, compare_torsiondrive_records
from qcportal.utils import now_at_utc
from .testing_helpers import (
    compare_torsiondrive_specs,
    test_specs,
    load_procedure_data,
    generate_task_key,
    load_record_data,
)

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal.managers import ManagerName
    from sqlalchemy.orm.session import Session


@pytest.mark.parametrize("spec", test_specs)
def test_torsiondrive_socket_add_get(
    storage_socket: SQLAlchemySocket, session: Session, spec: TorsiondriveSpecification
):
    hooh = load_molecule_data("peroxide2")
    td_mol_1 = load_molecule_data("td_C9H11NO2_1")
    td_mol_2 = load_molecule_data("td_C9H11NO2_2")

    time_0 = now_at_utc()
    meta, ids = storage_socket.records.torsiondrive.add(
        [[hooh], [td_mol_1, td_mol_2]], spec, True, "tag1", PriorityEnum.low, None, True
    )
    time_1 = now_at_utc()
    assert meta.success

    recs = [session.get(TorsiondriveRecordORM, i) for i in ids]

    assert len(recs) == 2
    for r in recs:
        assert r.record_type == "torsiondrive"
        assert r.status == RecordStatusEnum.waiting
        assert compare_torsiondrive_specs(spec, r.specification.model_dict())

        # Service queue entry should exist with the proper tag and priority
        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    assert len(recs[0].initial_molecules) == 1
    assert len(recs[1].initial_molecules) == 2

    assert recs[0].initial_molecules[0].molecule.identifiers["molecule_hash"] == hooh.get_hash()

    # Not necessarily in the input order
    hash1 = recs[1].initial_molecules[0].molecule.identifiers["molecule_hash"]
    hash2 = recs[1].initial_molecules[1].molecule.identifiers["molecule_hash"]
    assert {hash1, hash2} == {td_mol_1.get_hash(), td_mol_2.get_hash()}


def test_torsiondrive_socket_find_existing_1(storage_socket: SQLAlchemySocket):
    spec = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        ),
    )

    hooh = load_molecule_data("peroxide2")
    meta, id1 = storage_socket.records.torsiondrive.add([[hooh]], spec, True, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.torsiondrive.add([[hooh]], spec, True, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_torsiondrive_socket_find_existing_2(storage_socket: SQLAlchemySocket):
    # multiple molecule ordering, and duplicate molecules
    spec = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        ),
    )

    mol1 = load_molecule_data("td_C9H11NO2_0")
    mol2 = load_molecule_data("td_C9H11NO2_1")
    mol3 = load_molecule_data("td_C9H11NO2_2")
    meta, id1 = storage_socket.records.torsiondrive.add(
        [[mol1, mol2, mol3]], spec, True, "*", PriorityEnum.normal, None, True
    )
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.torsiondrive.add(
        [[mol2, mol3, mol1, mol2], [mol3, mol2, mol1, mol1]], spec, True, "*", PriorityEnum.normal, None, True
    )
    assert meta.n_inserted == 0
    assert meta.n_existing == 2
    assert meta.existing_idx == [0, 1]
    assert id2 == [id1[0], id1[0]]


def test_torsiondrive_socket_find_existing_3(storage_socket: SQLAlchemySocket):
    # some modifications to the input specification
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        ),
    )

    spec2 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optPROG1",
            keywords={"k": "value"},
            qc_specification=QCSpecification(
                program="prOG2",
                driver="deferred",
                method="b3LYP",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all", stdout=True),
            ),
        ),
    )

    mol1 = load_molecule_data("td_C9H11NO2_0")
    mol2 = load_molecule_data("td_C9H11NO2_1")
    mol3 = load_molecule_data("td_C9H11NO2_2")
    meta, id1 = storage_socket.records.torsiondrive.add(
        [[mol1, mol2, mol3]], spec1, True, "*", PriorityEnum.normal, None, True
    )
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.torsiondrive.add(
        [[mol1, mol2, mol3]], spec2, True, "*", PriorityEnum.normal, None, True
    )
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_torsiondrive_socket_add_different_1(storage_socket: SQLAlchemySocket):
    # Molecules are a subset of another
    spec = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        ),
    )

    mol1 = load_molecule_data("td_C9H11NO2_0")
    mol2 = load_molecule_data("td_C9H11NO2_1")
    mol3 = load_molecule_data("td_C9H11NO2_2")
    meta, id1 = storage_socket.records.torsiondrive.add(
        [[mol1, mol2, mol3]],
        spec,
        True,
        "*",
        PriorityEnum.normal,
        None,
        True,
    )
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.torsiondrive.add(
        [[mol1], [mol3, mol2], [mol2, mol3, mol1]],
        spec,
        True,
        "*",
        PriorityEnum.normal,
        None,
        True,
    )
    assert meta.n_inserted == 2
    assert meta.n_existing == 1
    assert meta.existing_idx == [2]
    assert meta.inserted_idx == [0, 1]
    assert id1[0] == id2[2]


@pytest.mark.parametrize(
    "test_data_name",
    [
        "td_C9H11NO2_mopac_pm6",
        "td_H2O2_mopac_pm6",
        "td_H2O2_psi4_pbe",
    ],
)
def test_torsiondrive_socket_run(
    storage_socket: SQLAlchemySocket, session: Session, activated_manager_name: ManagerName, test_data_name: str
):
    input_spec_1, molecules_1, result_data_1 = load_procedure_data(test_data_name)

    storage_socket.users.add(UserInfo(username="submit_user", role="submit", enabled=True))

    meta_1, id_1 = storage_socket.records.torsiondrive.add(
        [molecules_1], input_spec_1, True, "test_tag", PriorityEnum.low, "submit_user", True
    )
    id_1 = id_1[0]
    assert meta_1.success

    time_0 = now_at_utc()
    finished, n_optimizations = run_service(
        storage_socket, activated_manager_name, id_1, generate_task_key, result_data_1, 1000
    )
    time_1 = now_at_utc()

    rec = session.get(TorsiondriveRecordORM, id_1)

    assert rec.status == RecordStatusEnum.complete
    assert time_0 < rec.modified_on < time_1
    assert len(rec.compute_history) == 1
    assert len(rec.compute_history[-1].outputs) == 1
    assert rec.compute_history[-1].status == RecordStatusEnum.complete
    assert time_0 < rec.compute_history[-1].modified_on < time_1
    assert rec.service is None

    desc_info = storage_socket.records.get_short_descriptions([id_1])[0]
    short_desc = desc_info["description"]
    assert desc_info["record_type"] == rec.record_type
    assert desc_info["created_on"] == rec.created_on
    assert rec.specification.program in short_desc
    assert rec.specification.optimization_specification.program in short_desc
    assert rec.specification.optimization_specification.qc_specification.program in short_desc
    assert rec.specification.optimization_specification.qc_specification.method in short_desc

    out = rec.compute_history[-1].outputs["stdout"].get_output()
    assert "Job Finished" in out

    assert len(rec.optimizations) == n_optimizations


def test_torsiondrive_socket_run_duplicate(
    storage_socket: SQLAlchemySocket,
    session: Session,
    activated_manager_name: ManagerName,
):
    input_spec_1, molecules_1, result_data_1 = load_procedure_data("td_H2O2_mopac_pm6")

    meta_1, id_1 = storage_socket.records.torsiondrive.add(
        [molecules_1], input_spec_1, True, "test_tag", PriorityEnum.low, None, True
    )
    id_1 = id_1[0]
    assert meta_1.success

    run_service(storage_socket, activated_manager_name, id_1, generate_task_key, result_data_1, 1000)

    rec_1 = session.get(TorsiondriveRecordORM, id_1)
    assert rec_1.status == RecordStatusEnum.complete
    opt_ids_1 = [x.optimization_id for x in rec_1.optimizations]

    # Submit again, without duplicate checking
    meta_2, id_2 = storage_socket.records.torsiondrive.add(
        [molecules_1], input_spec_1, True, "test_tag", PriorityEnum.low, None, False
    )
    id_2 = id_2[0]
    assert meta_2.success
    assert id_2 != id_1

    run_service(storage_socket, activated_manager_name, id_2, generate_task_key, result_data_1, 1000)

    rec_2 = session.get(TorsiondriveRecordORM, id_2)
    assert rec_2.status == RecordStatusEnum.complete
    opt_ids_2 = [x.optimization_id for x in rec_2.optimizations]

    assert set(opt_ids_1).isdisjoint(opt_ids_2)


def test_torsiondrive_socket_insert_full_qcportal_record(secure_snowflake: QCATestingSnowflake):
    test_names = [
        "td_C9H11NO2_mopac_pm6",
        "td_H2O2_mopac_pm6",
        "td_H2O2_psi4_pbe0",
        "td_H2O2_psi4_pbe",
        "td_error_137174663",
    ]

    storage_socket = secure_snowflake.get_storage_socket()
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    user_id = client.get_user().id

    for test_name in test_names:
        initial_record = load_record_data(test_name)
        initial_record_copy = initial_record.copy(deep=True)

        # Need a full copy of results - they can get mutated
        with storage_socket.session_scope() as session:
            ins_ids = storage_socket.records.insert_full_qcportal_records(session, [initial_record_copy], user_id)

        rec_1 = client.get_torsiondrives(ins_ids[0], include=["**"])
        rec_1.fetch_children(include=["**"], force_fetch=True)
        compare_torsiondrive_records(rec_1, initial_record)

        assert rec_1.creator_user == "submit_user"
        for opts in rec_1.optimizations.values():
            for opt in opts:
                assert opt.creator_user == "submit_user"



from __future__ import annotations, annotations

from typing import TYPE_CHECKING

from qcarchivetesting import test_users
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake


def test_serverinfo_client_delete_error(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])
    storage_socket = secure_snowflake.get_storage_socket()

    admin_id = client.get_user("admin_user").id
    read_id = client.get_user("read_user").id

    error_data_1 = {
        "error_text": "This is a test error",
        "user_id": admin_id,
        "request_path": "/api/v1/molecule",
        "request_headers": "fake_headers",
        "request_body": "fake body",
    }

    error_data_2 = {
        "error_text": "This is another test error",
        "user_id": read_id,
        "request_path": "/api/v1/molecule",
        "request_headers": "fake_headers",
        "request_body": "fake body",
    }

    time_0 = now_at_utc()
    storage_socket.serverinfo.save_error(error_data_1)
    time_12 = now_at_utc()
    storage_socket.serverinfo.save_error(error_data_2)

    n_deleted = client.delete_error_log(before=time_0)
    assert n_deleted == 0

    n_deleted = client.delete_error_log(before=time_12)
    assert n_deleted == 1

    query_res = client.query_error_log()
    errors = list(query_res)
    assert len(errors) == 1
    assert errors[0].user == "read_user"



from flask import redirect, Response, stream_with_context, current_app

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route


@api_v1.route("/external_files/<int:file_id>", methods=["GET"])
@wrap_global_route("records", "read")
def get_external_file_metadata_v1(file_id: int):
    return storage_socket.external_files.get_metadata(file_id)


@api_v1.route("/external_files/<int:file_id>/download", methods=["GET"])
@wrap_global_route("records", "read")
def download_external_file_v1(file_id: int):
    passthrough = current_app.config["QCFRACTAL_CONFIG"].s3.passthrough

    if passthrough:
        file_name, streamer_func = storage_socket.external_files.get_file_streamer(file_id)

        return Response(
            stream_with_context(streamer_func()),
            content_type="application/octet-stream",
            headers={"Content-Disposition": f'attachment; filename="{file_name}"'},
        )
    else:
        _, url = storage_socket.external_files.get_url(file_id)
        return redirect(url, code=302)


@api_v1.route("/external_files/<int:file_id>/direct_link", methods=["GET"])
@wrap_global_route("records", "read")
def get_direct_link_external_file_v1(file_id: int):
    passthrough = current_app.config["QCFRACTAL_CONFIG"].s3.passthrough

    if passthrough:
        # Will check for missing file id
        meta = storage_socket.external_files.get_metadata(file_id)
        return f"/api/v1/external_files/{meta['id']}/download"
    else:
        _, url = storage_socket.external_files.get_url(file_id)
        return url



from __future__ import annotations

import logging
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from sqlalchemy import and_, update, select
from sqlalchemy.dialects.postgresql import array

from qcfractal.db_socket.helpers import get_query_proj_options, get_count, get_general
from qcportal.exceptions import ComputeManagerError
from qcportal.managers import ManagerStatusEnum, ManagerName, ManagerQueryFilters
from qcportal.utils import now_at_utc
from .db_models import ComputeManagerORM

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import List, Iterable, Optional, Sequence, Sequence, Dict, Any


class ManagerSocket:
    """
    Socket for managing/querying compute managers
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

        self._manager_heartbeat_frequency = root_socket.qcf_config.heartbeat_frequency
        self._manager_heartbeat_frequency_jitter = root_socket.qcf_config.heartbeat_frequency_jitter
        self._manager_max_missed_heartbeats = root_socket.qcf_config.heartbeat_max_missed

        with self.root_socket.session_scope() as session:
            self.root_socket.internal_jobs.add(
                "check_manager_heartbeats",
                now_at_utc(),
                "managers.check_manager_heartbeats",
                {},
                user_id=None,
                unique_name=True,
                repeat_delay=self._manager_heartbeat_frequency,
                session=session,
            )

    def activate(
        self,
        name_data: ManagerName,
        manager_version: str,
        username: Optional[str],
        programs: Dict[str, List[str]],
        compute_tags: List[str],
        *,
        session: Optional[Session] = None,
    ) -> int:
        """
        Activates a new manager on the server
        """

        # Strip out empty tags and programs
        compute_tags = [x.lower() for x in compute_tags if len(x) > 0]

        # Some version strings can contain uppercase characters
        programs = {k.lower(): [v.lower() for v in vlst] for k, vlst in programs.items() if len(k) > 0}

        if len(compute_tags) == 0:
            raise ComputeManagerError("Manager does not have any tags assigned. Use '*' to match all tags")
        if len(programs) == 0:
            raise ComputeManagerError("Manager does not have any programs available")

        compute_tags = list(dict.fromkeys(compute_tags))  # remove duplicates, maintaining order (in python 3.6+)

        manager_orm = ComputeManagerORM(
            name=name_data.fullname,
            cluster=name_data.cluster,
            hostname=name_data.hostname,
            username=username,
            compute_tags=compute_tags,
            status=ManagerStatusEnum.active,
            manager_version=manager_version,
            programs=programs,
        )

        with self.root_socket.optional_session(session) as session:
            stmt = select(ComputeManagerORM).where(ComputeManagerORM.name == name_data.fullname)
            count = get_count(session, stmt)

            if count > 0:
                self._logger.warning(f"Cannot activate duplicate manager: {name_data.fullname}")
                raise ComputeManagerError("A manager with this name already exists")

            session.add(manager_orm)
            session.flush()
            return manager_orm.id

    def update_resource_stats(
        self,
        name: str,
        active_tasks: int,
        active_cores: int,
        active_memory: float,
        total_cpu_hours: float,
        *,
        session: Optional[Session] = None,
    ):
        """
        Updates the resources available/in use by a manager
        """

        with self.root_socket.optional_session(session) as session:
            stmt = select(ComputeManagerORM).where(ComputeManagerORM.name == name).with_for_update(skip_locked=False)
            manager: Optional[ComputeManagerORM] = session.execute(stmt).scalar_one_or_none()

            if manager is None:
                raise ComputeManagerError(f"Cannot update resource stats for manager {name} - does not exist")
            if manager.status != ManagerStatusEnum.active:
                raise ComputeManagerError(f"Cannot update resource stats for manager {name} - is not active")

            manager.active_tasks = active_tasks
            manager.active_cores = active_cores
            manager.active_memory = active_memory
            manager.total_cpu_hours = total_cpu_hours
            manager.modified_on = now_at_utc()

    def deactivate(
        self,
        name: Optional[Iterable[str]] = None,
        modified_before: Optional[datetime] = None,
        *,
        reason: str = "(none given)",
        session: Optional[Session] = None,
    ) -> List[str]:
        """Marks managers as inactive

        If both name and modified_before are specified, managers that match both conditions will be deactivated.

        Parameters
        ----------
        name
            Names of managers to mark as inactive
        modified_before
            Mark all managers that were last modified before this date as inactive
        reason
            A descriptive reason given for deactivation
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of manager names that were marked as inactive

        """
        if not name and not modified_before:
            return []

        now = now_at_utc()
        query_and = []
        if name:
            query_and.append(ComputeManagerORM.name.in_(name))
        if modified_before:
            query_and.append(ComputeManagerORM.modified_on < modified_before)

        stmt = (
            update(ComputeManagerORM)
            .where(and_(ComputeManagerORM.status == ManagerStatusEnum.active, and_(True, *query_and)))
            .values(status=ManagerStatusEnum.inactive, modified_on=now)
            .returning(ComputeManagerORM.name)
        )

        with self.root_socket.optional_session(session) as session:
            deactivated_names = session.execute(stmt).fetchall()
            deactivated_names = [x[0] for x in deactivated_names]

            # For the manager, also reset any now-orphaned tasks that belonged to that manager
            for dead_name in deactivated_names:
                incomplete_ids = self.root_socket.records.reset_assigned(manager_name=[dead_name], session=session)
                self._logger.info(
                    f"Deactivated manager {dead_name}. Reason: {reason}. Recycling {len(incomplete_ids)} incomplete tasks."
                )

        return deactivated_names

    def get(
        self,
        name: Sequence[str],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        """
        Obtain manager information with specified names from the database

        Names for managers are unique, since they include a UUID.

        Parameters
        ----------
        name
            A list or other sequence of manager names
        include
            Which fields of the manager info to return. Default is to return all fields.
        exclude
            Remove these fields from the return. Default is to return all fields.
        missing_ok
           If set to True, then missing managers will be tolerated, and the returned list of
           managers will contain None for the corresponding IDs that were not found.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            List of manager data (as dictionaries) in the same order as the given names.
            If missing_ok is True, then this list will contain None where the manager was missing.
        """

        with self.root_socket.optional_session(session, True) as session:
            return get_general(session, ComputeManagerORM, ComputeManagerORM.name, name, include, exclude, missing_ok)

    def query(
        self,
        query_data: ManagerQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        """
        General query of managers in the database

        All search criteria are merged via 'and'. Therefore, records will only
        be found that match all the criteria.

        Parameters
        ----------
        query_data
            Fields/filters to query for
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of manager info (as dictionaries) that were found in the database.
        """

        proj_options = get_query_proj_options(ComputeManagerORM, query_data.include, query_data.exclude)

        and_query = []
        if query_data.manager_id is not None:
            and_query.append(ComputeManagerORM.id.in_(query_data.manager_id))
        if query_data.name is not None:
            and_query.append(ComputeManagerORM.name.in_(query_data.name))
        if query_data.hostname is not None:
            and_query.append(ComputeManagerORM.hostname.in_(query_data.hostname))
        if query_data.cluster is not None:
            and_query.append(ComputeManagerORM.cluster.in_(query_data.cluster))
        if query_data.status is not None:
            and_query.append(ComputeManagerORM.status.in_(query_data.status))
        if query_data.modified_before is not None:
            and_query.append(ComputeManagerORM.modified_on < query_data.modified_before)
        if query_data.modified_after is not None:
            and_query.append(ComputeManagerORM.modified_on > query_data.modified_after)

        with self.root_socket.optional_session(session, True) as session:
            stmt = select(ComputeManagerORM).filter(and_(True, *and_query))
            stmt = stmt.options(*proj_options)

            if query_data.cursor is not None:
                stmt = stmt.where(ComputeManagerORM.id < query_data.cursor)

            stmt = stmt.order_by(ComputeManagerORM.id.desc())
            stmt = stmt.limit(query_data.limit)
            stmt = stmt.distinct(ComputeManagerORM.id)

            results = session.execute(stmt).scalars().all()
            result_dicts = [x.model_dict() for x in results]

        return result_dicts

    def query_active(
        self, compute_tag: List[str], programs: Dict[str, List[str]], *, session: Optional[Session] = None
    ) -> List[str]:
        """
        Queries active managers for any that can take the given compute tag or program

        This will select managers that can handle any of the given compute tags, and all the given programs/versions.

        Parameters
        ----------
        compute_tag
            List of possible tags the task can have. A manager must have one of the tags to be considered available.
        programs
            Required programs/versions for the task. A manager must have all the programs/versions to be considered available.
        session
            An existing SQLAlchemy session to use.

        Returns
        -------
        :
            List of manager names that could possible handle the given compute tags/programs.
        """

        stmt = select(ComputeManagerORM.name)
        stmt = stmt.where(ComputeManagerORM.status == ManagerStatusEnum.active)

        # TODO - handle versions
        program_names = array(programs.keys())
        stmt = stmt.where(ComputeManagerORM.programs.op("?&")(program_names))  # ?& = JSONB contains the keys of ...

        # Handle *
        # If the server doesn't have strict_compute_tags and there is a * in the tags, then
        # don't limit the query
        if "*" not in compute_tag or ("*" in compute_tag and self.root_socket.qcf_config.strict_compute_tags):
            stmt = stmt.where(ComputeManagerORM.compute_tags.op("&&")(compute_tag))  # && = overlaps between two arrays

        with self.root_socket.optional_session(session, True) as session:
            manager_names = session.execute(stmt).scalars().all()
            return list(manager_names)

    def check_manager_heartbeats(self, session: Session) -> None:
        """
        Checks for manager heartbeats

        If a manager has not been heard from in a while, it is set to inactivate and its tasks
        reset to a waiting state. The amount of time to wait for a manager is controlled by the config
        options manager_max_missed_heartbeats and manager_heartbeat_frequency.

        Parameters
        ----------
        session
            An existing SQLAlchemy session to use.
        """
        self._logger.debug("Checking manager heartbeats")

        # Take into account the maximum jitter allowed
        manager_window = self._manager_max_missed_heartbeats * (
            self._manager_heartbeat_frequency * (1 + self._manager_heartbeat_frequency_jitter)
        )
        dt = now_at_utc() - timedelta(seconds=manager_window)

        dead_managers = self.deactivate(modified_before=dt, reason="missing heartbeat", session=session)

        if dead_managers:
            self._logger.info(f"Deactivated {len(dead_managers)} managers due to missing heartbeats")



###################################################
# The version stuff must be handled first.
# Other packages that we import later will need it
###################################################

import os
from importlib.metadata import version

__version__ = version("qcfractal")

# The location of this file
qcfractal_topdir = os.path.abspath(os.path.dirname(__file__))



from qcarchivetesting import load_hash_test_data
from qcfractal.components.manybody.record_db_models import ManybodySpecificationORM
from qcfractal.components.testing_fixtures import spec_test_runner
from qcfractal.db_socket import SQLAlchemySocket
from qcportal.manybody import ManybodySpecification
from qcportal.singlepoint import SinglepointProtocols, QCSpecification


def test_manybody_hash_canaries(storage_socket: SQLAlchemySocket):
    # Test data is hash : spec dict
    test_data = load_hash_test_data("manybody_specification_tests")

    spec_map = [(k, ManybodySpecification(**v)) for k, v in test_data.items()]

    specs = [x[1] for x in spec_map]
    meta, ids = storage_socket.records.manybody.add_specifications(specs)
    assert meta.success
    assert len(ids) == len(specs)
    assert meta.n_existing == 0

    with storage_socket.session_scope() as session:
        for spec_id, (spec_hash, _) in zip(ids, spec_map):
            spec_orm = session.get(ManybodySpecificationORM, spec_id)
            assert spec_orm.specification_hash == spec_hash


def test_manybody_socket_add_specification_same_1(spec_test_runner):
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["nocp"],
        keywords={"return_total_data": True},
    )

    spec_test_runner("manybody", spec1, spec1, True)


def test_manybody_socket_add_specification_same_2(spec_test_runner):
    # Test case sensitivity
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["nocp"],
        keywords={"return_total_data": True},
    )

    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prOg1",
                driver="energy",
                method="b3LYP",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["nocp"],
        keywords={"return_total_data": True},
    )

    spec_test_runner("manybody", spec1, spec2, True)


def test_manybody_socket_add_specification_same_3(spec_test_runner):
    # Test supersystem
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            "supersystem": QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["nocp"],
        keywords={"return_total_data": True},
    )

    spec_test_runner("manybody", spec1, spec1, True)


def test_manybody_socket_add_specification_same_4(spec_test_runner):
    # Test ordering
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="hf",
                basis="sto-3g",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            2: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            "supersystem": QCSpecification(
                program="prog1",
                driver="energy",
                method="ccsd",
                basis="aug-cc-pvtz",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["nocp"],
        keywords={},
    )

    # Test ordering
    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            2: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            "supersystem": QCSpecification(
                program="prog1",
                driver="energy",
                method="ccsd",
                basis="aug-cc-pvtz",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="hf",
                basis="sto-3g",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["nocp"],
        keywords={},
    )

    spec_test_runner("manybody", spec1, spec2, True)


def test_manybody_socket_add_specification_diff_1(spec_test_runner):
    # Test different parameters
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["nocp"],
        keywords={"return_total_data": True},
    )

    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["cp"],
        keywords={"return_total_data": True},
    )

    spec_test_runner("manybody", spec1, spec2, False)


def test_manybody_socket_add_specification_diff_2(spec_test_runner):
    # Test different parameters
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["nocp"],
        keywords={"return_total_data": True},
    )

    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["vmfc"],
        keywords={"return_total_data": True},
    )

    spec_test_runner("manybody", spec1, spec2, False)


def test_manybody_socket_add_specification_diff_3(spec_test_runner):
    # Test different parameters
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["cp"],
        keywords={"return_total_data": True},
    )

    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["cp"],
        keywords={"return_total_data": False},
    )
    spec_test_runner("manybody", spec1, spec2, False)


def test_manybody_socket_add_specification_diff_4(spec_test_runner):
    # Test different parameters
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["cp"],
        keywords={"return_total_data": True},
    )

    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            )
        },
        bsse_correction=["cp"],
    )
    spec_test_runner("manybody", spec1, spec2, False)


def test_manybody_socket_add_specification_diff_5(spec_test_runner):
    # Test different levels
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            2: QCSpecification(
                program="prog1",
                driver="energy",
                method="hf",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["cp"],
        keywords={},
    )

    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["cp"],
        keywords={},
    )

    spec_test_runner("manybody", spec1, spec2, False)


def test_manybody_socket_add_specification_diff_6(spec_test_runner):
    # Test different levels
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="ccsd",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            2: QCSpecification(
                program="prog1",
                driver="energy",
                method="hf",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["cp"],
        keywords={"return_total_data": True},
    )

    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            2: QCSpecification(
                program="prog1",
                driver="energy",
                method="hf",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["cp"],
        keywords={"return_total_data": True},
    )

    spec_test_runner("manybody", spec1, spec2, False)


def test_manybody_socket_add_specification_diff_7(spec_test_runner):
    # Test different levels
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            2: QCSpecification(
                program="prog1",
                driver="energy",
                method="hf",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["cp"],
        keywords={"return_total_data": True},
    )

    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            2: QCSpecification(
                program="prog1",
                driver="energy",
                method="hf",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="none"),
            ),
        },
        bsse_correction=["cp"],
        keywords={"return_total_data": True},
    )
    spec_test_runner("manybody", spec1, spec2, False)


def test_manybody_socket_add_specification_diff_8(spec_test_runner):
    # Test different levels
    spec1 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="ccsd",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            2: QCSpecification(
                program="prog1",
                driver="energy",
                method="hf",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["cp"],
    )

    spec2 = ManybodySpecification(
        program="qcmanybody",
        levels={
            1: QCSpecification(
                program="prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
            "supersystem": QCSpecification(
                program="prog1",
                driver="energy",
                method="hf",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        bsse_correction=["cp"],
    )

    spec_test_runner("manybody", spec1, spec2, False)



"""add user session table

Revision ID: 521464fa93ca
Revises: 2c3c406a2750
Create Date: 2025-03-23 16:17:36.143307

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "521464fa93ca"
down_revision = "2c3c406a2750"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "user_session",
        sa.Column("session_id", sa.String(), nullable=False),
        sa.Column("session_data", sa.LargeBinary(), nullable=False),
        sa.Column("last_accessed", sa.TIMESTAMP(timezone=True), nullable=False),
        sa.PrimaryKeyConstraint("session_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("user_session")
    # ### end Alembic commands ###



"""
Utility functions for handling network ports
"""

import socket
from typing import Optional


def find_open_port(host: str = "localhost", starting_port: Optional[int] = None) -> int:
    """
    Finds an open port that we can bind to
    """

    if starting_port is None:
        sock = socket.socket()
        sock.bind((host, 0))
        _, port = sock.getsockname()
    else:
        port = starting_port
        while is_port_inuse(host, port):
            port += 1

    return port


def is_port_inuse(host: str, port: int) -> bool:
    """
    Determine if an ip/port is being used or not

    Parameters
    ----------
    host
        The host name or IP address to check
    port
        The port on the host to check

    Returns
    -------
    bool
        True if the port is currently in use, False otherwise
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(1)
        try:
            s.connect((host, int(port)))
            s.shutdown(socket.SHUT_RDWR)
            return True
        except ConnectionRefusedError:
            return False



"""Rename tag and priority columns

Revision ID: 2abde51b4bc7
Revises: 236316a77635
Create Date: 2025-03-14 12:58:18.463114

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2abde51b4bc7"
down_revision = "236316a77635"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("record_info_backup", "old_tag", new_column_name="old_compute_tag")
    op.alter_column("record_info_backup", "old_priority", new_column_name="old_compute_priority")

    op.alter_column("service_queue", "tag", new_column_name="compute_tag")
    op.alter_column("service_queue", "priority", new_column_name="compute_priority")
    op.drop_index("ix_service_queue_tag", table_name="service_queue")
    op.create_index("ix_service_queue_compute_tag", "service_queue", ["compute_tag"], unique=False)

    op.alter_column("task_queue", "tag", new_column_name="compute_tag")
    op.alter_column("task_queue", "priority", new_column_name="compute_priority")
    op.drop_index("ix_task_queue_sort", table_name="task_queue", postgresql_where="(available = true)")
    op.create_index(
        "ix_task_queue_sort",
        "task_queue",
        [sa.text("compute_priority DESC"), sa.text("sort_date ASC"), sa.text("id ASC"), "compute_tag"],
        unique=False,
        postgresql_where=sa.text("available = true"),
    )
    op.drop_index("ix_task_queue_tag", table_name="task_queue")
    op.create_index("ix_task_queue_tag", "task_queue", ["compute_tag"], unique=False)

    op.drop_constraint("ck_task_queue_tag_lower", "task_queue")
    op.create_check_constraint(
        "ck_task_queue_compute_tag_lower",
        "task_queue",
        sa.column("compute_tag").cast(sa.TEXT) == sa.func.lower(sa.column("compute_tag").cast(sa.TEXT)),
    )

    op.drop_constraint("ck_service_queue_tag_lower", "service_queue")
    op.create_check_constraint(
        "ck_service_queue_compute_tag_lower",
        "service_queue",
        sa.column("compute_tag").cast(sa.TEXT) == sa.func.lower(sa.column("compute_tag").cast(sa.TEXT)),
    )

    op.alter_column("base_dataset", "default_tag", new_column_name="default_compute_tag")
    op.alter_column("base_dataset", "default_priority", new_column_name="default_compute_priority")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("record_info_backup", "old_compute_tag", new_column_name="old_tag")
    op.alter_column("record_info_backup", "old_compute_priority", new_column_name="old_priority")

    op.alter_column("service_queue", "compute_tag", new_column_name="tag")
    op.alter_column("service_queue", "compute_priority", new_column_name="priority")
    op.drop_index("ix_service_queue_compute_tag", table_name="service_queue")
    op.create_index("ix_service_queue_tag", "service_queue", ["tag"], unique=False)

    op.alter_column("task_queue", "compute_tag", new_column_name="tag")
    op.alter_column("task_queue", "compute_priority", new_column_name="priority")
    op.drop_index("ix_task_queue_tag", table_name="task_queue")
    op.create_index("ix_task_queue_tag", "task_queue", ["tag"], unique=False)
    op.drop_index("ix_task_queue_sort", table_name="task_queue", postgresql_where=sa.text("available = true"))
    op.create_index(
        "ix_task_queue_sort",
        "task_queue",
        [sa.text("priority DESC"), "sort_date", "id", "tag"],
        unique=False,
        postgresql_where="(available = true)",
    )

    op.drop_constraint("ck_task_queue_compute_tag_lower", "task_queue")
    op.create_check_constraint(
        "ck_task_queue_tag_lower",
        "task_queue",
        sa.column("tag").cast(sa.TEXT) == sa.func.lower(sa.column("tag").cast(sa.TEXT)),
    )

    op.drop_constraint("ck_service_queue_compute_tag_lower", "service_queue")
    op.create_check_constraint(
        "ck_service_queue_tag_lower",
        "service_queue",
        sa.column("tag").cast(sa.TEXT) == sa.func.lower(sa.column("tag").cast(sa.TEXT)),
    )

    op.alter_column("base_dataset", "default_compute_tag", new_column_name="default_tag")
    op.alter_column("base_dataset", "default_compute_priority", new_column_name="default_priority")
    # ### end Alembic commands ###



from typing import List

from flask import current_app, g

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.exceptions import LimitExceededError
from qcportal.neb import NEBDatasetSpecification, NEBDatasetNewEntry, NEBAddBody, NEBQueryFilters
from qcportal.utils import calculate_limit


#####################
# Record
#####################


@api_v1.route("/records/neb/bulkCreate", methods=["POST"])
@wrap_global_route("records", "add")
def add_neb_records_v1(body_data: NEBAddBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.add_records
    if len(body_data.initial_chains) > limit:
        raise LimitExceededError(f"Cannot add {len(body_data.initial_chains)} neb records - limit is {limit}")

    return storage_socket.records.neb.add(
        initial_chains=body_data.initial_chains,
        neb_spec=body_data.specification,
        compute_tag=body_data.compute_tag,
        compute_priority=body_data.compute_priority,
        creator_user=g.username,
        find_existing=body_data.find_existing,
    )


@api_v1.route("/records/neb/<int:record_id>/neb_result", methods=["GET"])
@wrap_global_route("records", "read")
def get_neb_result_v1(record_id: int):
    return storage_socket.records.neb.get_neb_result(record_id)


@api_v1.route("/records/neb/<int:record_id>/singlepoints", methods=["GET"])
@wrap_global_route("records", "read")
def get_neb_singlepoints_v1(record_id: int):
    return storage_socket.records.neb.get_singlepoints(record_id)


@api_v1.route("/records/neb/<int:record_id>/optimizations", methods=["GET"])
@wrap_global_route("records", "read")
def get_neb_optimizations_v1(record_id: int):
    return storage_socket.records.neb.get_optimizations(record_id)


@api_v1.route("/records/neb/<int:record_id>/initial_chain", methods=["GET"])
@wrap_global_route("records", "read")
def get_neb_initial_chain_molecule_ids_v1(record_id: int):
    return storage_socket.records.neb.get_initial_molecules_ids(record_id)


@api_v1.route("/records/neb/query", methods=["POST"])
@wrap_global_route("records", "read")
def query_neb_v1(body_data: NEBQueryFilters):
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_records
    body_data.limit = calculate_limit(max_limit, body_data.limit)

    return storage_socket.records.neb.query(body_data)


#####################
# Dataset
#####################


@api_v1.route("/datasets/neb/<int:dataset_id>/specifications", methods=["POST"])
@wrap_global_route("datasets", "modify")
def add_neb_dataset_specifications_v1(dataset_id: int, body_data: List[NEBDatasetSpecification]):
    return storage_socket.datasets.neb.add_specifications(dataset_id, body_data)


@api_v1.route("/datasets/neb/<int:dataset_id>/entries/bulkCreate", methods=["POST"])
@wrap_global_route("datasets", "modify")
def add_neb_dataset_entries_v1(dataset_id: int, body_data: List[NEBDatasetNewEntry]):
    return storage_socket.datasets.neb.add_entries(dataset_id, new_entries=body_data)


@api_v1.route("/datasets/neb/<int:dataset_id>/background_add_entries", methods=["POST"])
@wrap_global_route("datasets", "modify")
def background_add_neb_dataset_entries_v1(dataset_id: int, body_data: List[NEBDatasetNewEntry]):
    return storage_socket.datasets.neb.background_add_entries(dataset_id, new_entries=body_data)



from __future__ import annotations

from queue import Queue
from typing import Optional
from weakref import WeakKeyDictionary

from flask import Flask

from qcfractal.db_socket import SQLAlchemySocket


class FlaskStorageSocket:

    _app_sockets: WeakKeyDictionary[Flask, SQLAlchemySocket]

    def __init__(self):
        self._app_sockets = WeakKeyDictionary()

    def init_app(self, app: Flask, finished_queue: Optional[Queue] = None):
        socket = SQLAlchemySocket(app.config["QCFRACTAL_CONFIG"])

        if not hasattr(app, "extensions"):
            app.extensions = {}

        app.extensions["storage_socket"] = socket

        if finished_queue:
            socket.set_finished_watch(finished_queue)

        self._app_sockets[app] = socket

    def get_socket(self, app) -> SQLAlchemySocket:
        s = self._app_sockets.get(app, None)
        if s is None:
            raise RuntimeError("Socket not initialized for this flask app")

        return s



import time

import pytest

from qcfractal.snowflake import FractalSnowflake


@pytest.mark.slow
def test_snowflake_restarting(tmp_path):
    s = FractalSnowflake(start=False, tmpdir_parent=str(tmp_path))

    s._start_api()
    s._start_compute()
    s._start_job_runner()

    time.sleep(5)

    s.stop()

    time.sleep(5)

    assert s._api_thread is None
    assert s._compute_thread is None
    assert s._job_runner_thread is None

    s._start_api()
    s._start_compute()
    s._start_job_runner()

    assert s._api_thread.is_alive()
    assert s._compute_thread.is_alive()
    assert s._job_runner_thread.is_alive()

    c = s.client()
    assert c.ping()
